"""
iogp.data.str: Strings.

Author: Vlad Topan (vtopan/gmail)
"""
import ast
import codecs
import datetime
from logging import error
import math
import re
import time

from .map437 import CP437_TO_UNICODE, UNICODE_TO_CP437


NUM_MODS = 'KMBTQ'


def ellipsis(s, size):
    """
    Trim a string to a maximum size, add an ellipsis ("[...]") if trimmed.
    """
    if len(s) > size:
        s = s[:size - 5] + '[...]'
    return s


def parse_num(s):
    """
    Extract the value from a formatted number.
    """
    if s in ('N/A', None, '-', ''):
        return 0    # None
    if s == 0:
        return 0
    if re.search(r'^-?[,\.\d]+%$', s):
        # percent => normalize to 0.0..1.0
        return ast.literal_eval(s.replace(',', '').rstrip('%')) / 100
    s = s.replace(',', '')
    sign = 1
    if s[0] == '(' and s[-1] == ')':
        sign = -1
        s = s.strip('()')
    if s[-1] in NUM_MODS:
        value = int(float(s[:-1]) * (10 ** (3 * (NUM_MODS.index(s[-1]) + 1))))
    elif s[-1] in NUM_MODS.lower():
        value = int(float(s[:-1]) * (10 ** (3 * (NUM_MODS.lower().index(s[-1]) + 1))))
    else:
        try:
            value = float(s) if '.' in s else int(s)
        except ValueError:
            error(f'Failed parsing {s} as float/int!')
            value = 0
    value = sign * value
    return value


def parse_percent(text):
    """
    Extract a percent float (0.0 .. 1.0) from text (with or without %).
    """
    return float(text.strip().strip('%')) / 100


def format_size(size):
    """
    Format a size in bytes.
    """
    for i, c in enumerate(('P', 'T', 'G', 'M', 'K', '')):
        v = (2 ** 10) ** (5 - i)
        if v <= size:
            s = str(int(size * 100 / v))
            s = s[:-2] if s.endswith('00') else s[:-2] + '.' + s[-2:]
            return s + f' {c}B'
    return '0B'


def format_num(num):
    """
    Format a (potentially large) number.
    """
    neg, num = int(num < 0), abs(num)
    for i, c in enumerate(list(reversed(NUM_MODS)) + ['']):
        v = 1000 ** (len(NUM_MODS) - i)
        if v <= num:
            s = str(int(num * 100 / v))
            s = s[:-2] if s.endswith('00') else s[:-2] + '.' + s[-2:]
            return (neg * '-') + s + c
    return f'{num:.3f}'


def format_date(ts):
    """
    Format a date, datetime or timestamp as a date.
    """
    if type(ts) in (datetime.datetime, datetime.date):
        return ts.strftime('%Y.%m.%d')
    if type(ts) in (int, float):
        ts = time.localtime(ts)
    return time.strftime('%Y.%m.%d', ts)


def format_datetime(ts):
    """
    Format a datetime or timestamp as a date + time (no seconds).
    """
    if type(ts) is datetime.datetime:
        return ts.strftime('%H:%M, %Y.%m.%d')
    if type(ts) in (int, float):
        ts = time.localtime(ts)
    return time.strftime('%H:%M, %Y.%m.%d', ts)


def format_percent(v):
    """
    Format a value as a percent.
    """
    return f'{(v or 0) * 100:.2f}%'


def int_to_bytes(val, size=None, byteorder='big'):
    """
    Convert an integer to a `bytes` string.

    This does the same thing as PyCryptodome's Crypto.Util.Number.long_to_bytes().

    :param byteorder: 'big' or 'little'.
    """
    size = size or int(math.ceil((val.bit_length() + 7) >> 3))
    res = val.to_bytes(size, byteorder=byteorder)
    return res


def bytes_to_int(val, byteorder='big'):
    """
    Convert a `bytes` string to an integer.

    This does the same thing as PyCryptodome's Crypto.Util.Number.bytes_to_long().

    :param byteorder: 'big' or 'little'.
    """
    res = int.from_bytes(val, byteorder=byteorder)
    return res


def dump_dict(title, d, indent=0):
    """
    Pretty-print dictionary whose values are other dicts, lists or basic types.
    """
    lines = []
    if title:
        lines.append(f'{indent * "  "}- {title}:')
    for k, v in d.items():
        if type(v) is dict:
            lines.append(dump_dict(k, v, indent + 1))
        elif type(v) is list:
            lines.append(f'{indent * "  "}  - {k}:')
            for i, e in enumerate(v):
                lines.append(dump_dict(f'#{i}', e, indent + 2))
        else:
            lines.append(f'{indent * "  "}  - {k}: {v}')
    return '\n'.join(lines)


def to_hex(data, wrap=None, truncate=None, sep=' '):
    """
    Encode `bytes` (or `str`) string as hex.

    :param wrap: Wrap at this many characters (rounded to 3).
    :param truncate: If longer than this many characters, truncate and append an ellipsis.
    """
    if isinstance(data, str):
        data = data.encode('ascii')
    data = data.hex()
    data = sep.join(a + b for a, b in zip(data[::2], data[1::2]))
    if truncate and len(data) > truncate:
        return data[:truncate] + '[...]'
    mul = 2 + len(sep)
    if wrap:
        wrap = wrap // mul * mul
        data = '\n'.join(data[i:i + wrap].rstrip(' ') for i in range(0, len(data), wrap))
    return data


def decode_cp437(data: bytes):
    """
    Decode bytes to str based on the CP437 DOS font / codepage.
    """
    return ''.join(CP437_TO_UNICODE[e] for e in data)


def clean_filename(filename, rep='_'):
    """
    Remove all potentially problematic symbols from a filename.
    """
    filename = re.sub(r'[^a-z0-9_\., ~@()=+-]+', rep, filename, flags=re.I).strip().rstrip(' .')
    return filename



"""
iogp.gui.qtwidgets: Custom Qt widgets.

Author: Vlad Topan (vtopan/gmail)
"""
import html
import threading
import time

from .qt import (QTextBrowser, QFontMetrics, QLineEdit, QTextEdit, QHBoxLayout, QLabel, QComboBox)
from .qthexed import VHexEditor, VBinViewer



class VMessageLog(QTextBrowser):

    def __init__(self, *args, rows=4, hint='Log messages', log_timestamps=False,
                open_external_links=True, **kwargs):
        # todo: more flexible style control
        super().__init__(*args, **kwargs)
        self.setStyleSheet('QScrollBar:vertical {width: 10px;}')
        self.setPlaceholderText(hint)
        self.setReadOnly(1)
        if rows:
            rowheight = QFontMetrics(self.font()).lineSpacing()
            self.setFixedHeight(10 + rows * rowheight)
        self.setOpenExternalLinks(open_external_links)
        self.lock = threading.Lock()
        self.log_timestamps = log_timestamps

    def log(self, msg, escape=True):
        """
        Log a message.
        """
        self.lock.acquire()
        if escape:
            msg = html.escape(msg)
        if msg.startswith('[DEBUG]'):
            msg = f'<font color="#444">{msg}</font>'
        elif msg.startswith('[ERROR]') or msg.startswith('[CRASH]'):
            msg = f'<font color="#A00">{msg}</font>'
        if self.log_timestamps:
            msg = f'<font color="#777">{time.strftime("[%d.%m.%y %H:%M:%S]")}</font> {msg}'
        if escape:
            msg = msg.replace('\n', '<br>')
        self.append(msg)
        self.ensureCursorVisible()
        self.lock.release()



def VLabelEdit(label, value='', placeholder='', validator=None, cls=QLineEdit, hint=None):
    """
    Returns a horizontal layout containing a label and an edit.

    The label and layout widgets are accessible as the `.label` / `.layout` attributes of the
    edit widget.
    """
    obj = cls()
    obj.label = QLabel(label)
    lay = obj.parent_layout = QHBoxLayout()
    lay.addWidget(obj.label)
    lay.addWidget(obj)
    obj.label.setBuddy(obj)
    if validator:
        obj.setValidator(validator)
    if isinstance(obj, QLineEdit):
        lay.text = obj.text
        lay.setText = obj.setText
    elif isinstance(obj, QTextEdit):
        lay.text = obj.toPlainText
        lay.setText = obj.setPlainText
    elif isinstance(obj, QComboBox):
        lay.text = obj.currentText
        lay.setText = obj.setCurrentText
    if value:
        lay.setText(value)
    if hint:
        obj.label.setToolTip(hint)
        obj.setToolTip(hint)
    if placeholder:
        obj.setPlaceholderText(placeholder)
    return obj



"""
iogp.db.cache: Generic DB-based URI caching objects.

Author: Vlad Topan (vtopan/gmail)
"""

from .sqlite import DB

import dbm
import os
import sys
import pickle
import marshal
import zlib
import time
import collections
try:
    import yaml
except ImportError:
    yaml = None
import json


DEF_DB_FILENAME = '.obj_cache.db'
DEF_MAX_AGE = 12 * 3600

ENC_MODE_ZLIB = 0x0001
ENC_MODE_PICKLE = 0x0002
ENC_MODE_MARSHAL = 0x0004
ENC_MODE_YAML = 0x0008
ENC_MODE_JSON = 0x0010
ENC_MODE_PZ = ENC_MODE_PICKLE | ENC_MODE_ZLIB
ENC_MODE_MARZ = ENC_MODE_MARSHAL | ENC_MODE_ZLIB
ENC_MODE_YZ = ENC_MODE_YAML | ENC_MODE_ZLIB
ENC_MODE_JZ = ENC_MODE_JSON | ENC_MODE_ZLIB

VALID_ENCODING_MODES = [
    ENC_MODE_ZLIB, ENC_MODE_PZ, ENC_MODE_PICKLE, ENC_MODE_YAML,
    ENC_MODE_MARSHAL, ENC_MODE_MARZ, ENC_MODE_YZ, ENC_MODE_JZ, ENC_MODE_JSON
    ]

CACHE_ALWAYS = -1  # always cache
CACHE_ONLY = -2  # only return from cache

BE_SQLITE = 1
BE_DBM = 2
BACKENDS = {
    'sqlite': BE_SQLITE,
    'dbm': BE_DBM,
    }


TABLES = {
    'cache': 'uri TEXT PRIMARY KEY, data BLOB, ts INT, enc_mode INT',
    }
INDEXES = {
    'cache': ['uri'],
    }


def ObjCache(filename=None, timing=0, **kargs):
    """
    Wrapper over DataCache which generates an automatic filename based on the main script name.

    Pass max_age=None to prevent items from ever expiring.
    """
    if not filename:
        filename = '_%s_obj.cache' % (os.path.basename(sys.argv[0])).split('.', 1)[0][:12]
    return DataCache(filename, timing=timing, **kargs)



class DataCache(object):
    """
    Generic URI-based data/object caching, backed by dbm or an SQLITE db.

    Can be used as a dictionary.

    :param encoding: Used for all non-string data; must be one of the ENC_MODE_* constants.
    :param max_age: 0 or negative value means "don't expire", otherwise number of seconds until entry expires.
    :param backend: 'sqlite' or 'dbm'.
    :param timing: Set to true to enable timing operations for debug purposes.
    """

    def __init__(self, filename=None, max_age=DEF_MAX_AGE, encoding=ENC_MODE_MARZ, timing=0, backend='sqlite',
                **kwargs):
        self.backend = BACKENDS[backend]
        self.encoding = kwargs.pop('obj_enc_mode', encoding)
        self.max_age = max_age
        self.filename = filename or DEF_DB_FILENAME
        self._db = None
        self._kwargs = kwargs
        self._connected = False
        self.timing = timing
        self.times = collections.defaultdict(lambda:list((0, 0)))
        self._del_count = 0


    @property
    def db(self):
        """
        Database object.
        """
        if not self._connected:
            if os.path.dirname(self.filename) and not os.path.isdir(os.path.dirname(self.filename)):
                os.makedirs(os.path.dirname(self.filename))
            # print('Connecting to DB [%s]...' % self.filename)
            if self.backend == BE_SQLITE:
                self._db = DB(self.filename, **self._kwargs)
                self._db.create_tables(TABLES)
                self._db.create_indexes(INDEXES)
            else:
                self._db = dbm.open(self.filename, 'c', **self._kwargs)
            self._connected = True
        return self._db


    def exists(self, uri):
        """
        Checks if a URI is present in the cache.
        """
        if self.backend == BE_SQLITE:
            res = self.db.fetchone("SELECT ts FROM cache WHERE uri = ?", (uri,))
        else:
            res = self.db.get(uri, None)
        if res and self.max_age and self.max_age > 0 and (time.time() - res[0] > self.max_age):
            res = None
        return bool(res)


    def get(self, uri, max_age=None):
        """
        Get cached data for the given URI.
        """
        t0 = time.time()
        if max_age is None:
            max_age = self.max_age
        if self.backend == BE_SQLITE:
            res = self.db.fetchone("SELECT data, ts, enc_mode FROM cache WHERE uri = ?", (uri,))
            if res:
                res, ts, enc = res
        else:
            res, enc = self.db.get(uri, None), self.encoding
        if res:
            try:
                res = decode_data(res, enc)
            except Exception as e:
                # data corruption
                print('[!] Corrupted data for URI [%s]: %s!' % (uri, e))
                # del self[uri]
                res = None
            if self.backend == BE_DBM:
                ts, res = res[1], res[0]
        if res and max_age and max_age > 0 and time.time() - ts > max_age:
            res = None
        if self.timing:
            self.times['get'][0] += 1
            self.times['get'][1] += time.time() - t0
        return res


    def get_age(self, uri):
        """
        Returns the age (in seconds) of an URI if present in the cache (or None).
        """
        if self.backend == BE_SQLITE:
            res = self.db.fetchone("SELECT ts FROM cache WHERE uri = ?", (uri,))
        else:
            res = self.db.get(uri, None)
        return time.time() - res[0] if res else None


    def __contains__(self, uri):
        return self.exists(uri)


    def __getitem__(self, uri, *args):
        res = self.get(uri)
        return res


    def __delitem__(self, uri):
        self._del_count += 1
        if self.backend == BE_SQLITE:
            self.db.execute("DELETE FROM cache WHERE uri = ?", (uri,))
        else:
            del self.db[uri]


    def __setitem__(self, uri, data):
        t0 = time.time()
        enc_mode = self.encoding
        if self.backend == BE_SQLITE:
            del self[uri]
            data = encode_data(data, enc_mode)
            try:
                self.db.execute("INSERT INTO cache(uri, ts, data, enc_mode) VALUES (?, ?, ?, ?)",
                    (uri, int(time.time()), data, enc_mode))
            except Exception as e:
                print(f'!!! insert failed for URI {uri}: {e}')
                raise
        else:
            self.db[uri] = encode_data([data, int(time.time()), enc_mode], enc_mode)
        if self.timing:
            self.times['set'][0] += 1
            self.times['set'][1] += time.time() - t0


    def __iter__(self):
        if self.backend == BE_SQLITE:
            self.__crt = None
            return self
        else:
            return self.db.__iter__()


    def __next__(self):
        return self.next()


    def next(self):
        """
        Iterator helper.
        """
        if self.__crt is None:
            # first item
            self.__crt = self.db.fetchvalue("SELECT uri FROM cache ORDER BY uri ASC LIMIT 1")
        else:
            # next
            self.__crt = self.db.fetchvalue("SELECT uri FROM cache WHERE uri > ? ORDER BY uri ASC LIMIT 1",
                    (self.__crt,))
        if not self.__crt:
            raise StopIteration()
        return self.__crt


    def close(self, vacuum=0):
        """
        Close DB connections (if any).
        """
        if self.backend == BE_SQLITE and vacuum and self._del_count:
            t0 = time.time()
            self.db.vacuum()
            if self.timing:
                self.times['vacuum'][0] += 1
                self.times['vacuum'][1] += time.time() - t0


    def dump_timing(self):
        """
        Print timing (debug) information.
        """
        if not self.times:
            print('No cache timing data!')
            return
        print('Cache timing data:')
        for k, v in self.times.items():
            print('%-8s %6d %7.2fs' % (k, v[0], v[1]))


    def keys(self):
        if self.backend == BE_DBM:
            return self.db.keys()
        return [x[0] for x in self.db.fetchall("SELECT uri FROM cache ORDER BY uri ASC")]


    def __len__(self):
        if self.backend == BE_DBM:
            return len(self.db)
        return self.db.fetchone("SELECT count(1) FROM cache")[0]


    def items(self):
        for k in self.keys():
            yield (k, self[k])



class DataStore(DataCache):
    """
    Data store which Caches part of the data in memory.
    """

    def __init__(self, filename=None, mcnt=10000, free=0.01, max_age=0, **kwargs):
        max_age = kwargs.pop('max_age', max_age)
        super().__init__(filename, max_age=max_age, **kwargs)
        self.mcnt = mcnt
        self.free = free if type(free) == int else int(mcnt * free)
        self._mcache = {}


    def get(self, uri, max_age=None):
        """
        Overloads `DataCache.get()` to implement in-memory caching.
        """
        if uri in self._mcache:
            e = self._mcache[uri]
            if not max_age or (time.time() - e[1] < max_age):
                return e[0]
        res = super().get(uri, max_age=max_age)
        self._mcache[uri] = [res, time.time()]
        if len(self._mcache) >= self.mcnt:
            remove = sorted(self._mcache.items(), key=lambda e:e[1][1])[:self.free]
            # print('RM: %d' % len(remove))
            for k, v in remove:
                del self._mcache[k]
        return res


    def __setitem__(self, uri, data):
        self._mcache[uri] = [data, time.time()]
        super().__setitem__(uri, data)



def encode_data(data, enc_mode):
    """
    Encodes and/or compresses data for serialization.

    :param data: `bytes` buffer.
    :param enc_mode: one or more `ENC_MODE_*` flags.
    """
    if enc_mode not in VALID_ENCODING_MODES:
        raise ValueError('Invalid encoding mode [%s]!' % enc_mode)
    if enc_mode & ENC_MODE_YAML:
        data = yaml.dumps(data)
    if enc_mode & ENC_MODE_JSON:
        data = json.dumps(data).encode('ascii')
    if enc_mode & ENC_MODE_PICKLE:
        data = pickle.dumps(data)
    if enc_mode & ENC_MODE_MARSHAL:
        data = marshal.dumps(data)
    if enc_mode & ENC_MODE_ZLIB:
        data = zlib.compress(data, level=zlib.Z_BEST_SPEED)
    return data


def decode_data(data, enc_mode):
    """
    Decodes and/or decompresses data for deserialization.

    :param data: `bytes` buffer.
    :param enc_mode: one or more `ENC_MODE_*` flags (must match the flags passed to `encode_data`).
    """
    if enc_mode not in VALID_ENCODING_MODES:
        raise ValueError('Invalid encoding mode [%s]!' % enc_mode)
    if enc_mode & ENC_MODE_ZLIB:
        data = zlib.decompress(data)
    if enc_mode & ENC_MODE_JSON:
        data = json.loads(data.decode('ascii'))
    if enc_mode & ENC_MODE_PICKLE:
        data = pickle.loads(data)
    if enc_mode & ENC_MODE_MARSHAL:
        data = marshal.loads(data)
    return data




"""
iogp.db.dbobj: Generic DB object wrapper (to be inherited by specific DB wrappers such as sqlite).

Author: Vlad Topan (vtopan/gmail)
"""

import inspect
import time
import sys
import threading


# in queries
PARAM = '?'
DEF_ID_KEY_PAT = '${KEY}'



class GenDBObj(object):
    """
    Generic DB (conn+cursor) object.

    :param log_all: Log all queries.
    :param cacheable: List of table names whose entries are cached.
    """

    def __init__(self, warn_long_time=2, time_queries=0, debug=0, log_all=0, cacheable=None,
                serialized=False, **kwarg):
        self.warn_long_time = warn_long_time
        self.log_all = log_all
        self.time_queries = time_queries
        self.cache = {}
        self.qtimes = {}
        self.debug = debug
        self.conn = self.cur = None
        self.serialized = serialized
        self.lock = threading.Lock()
        self.cacheable = cacheable if cacheable is not None else []


    def commit(self):
        if self.conn:
            self.conn.commit()


    def execute(self, query, args=None, log=0):
        if self.serialized:
            self.lock.acquire()
        query = query.replace(PARAM, self.ISYM).replace(DEF_ID_KEY_PAT, self.DEF_ID_KEY)
        if log or self.log_all:
            self.log('[DBG] QUERY [%s]; ARGS [%s]\n' % (query, args))
        t0 = time.time()
        try:
            self.cur.execute(query, args or [])
        except:     # fixme!
            self.log("QUERY [%s]; ARGS [%s] failed!" % (query, args))
            if self.serialized:
                self.lock.release()
            raise
        delta = time.time() - t0
        if self.warn_long_time and delta > self.warn_long_time:
            self.log('[DBG] Query [%s] took %5.2fs\n' % (query, delta), 'warn')
        if self.time_queries:
            k = '%-20s %s' % (caller()[:20], query[:80])
            if k not in self.qtimes:
                self.qtimes[k] = [1, delta]
            else:
                self.qtimes[k][0] += 1
                self.qtimes[k][1] += delta
        if self.serialized:
            self.lock.release()


    def executemany(self, query, args):
        if self.serialized:
            self.lock.acquire()
        self.conn.executemany(query, args)
        if self.serialized:
            self.lock.release()


    def fetchall(self, query, args=None, log=0, **kargs):
        self.execute(query, args, log=log, **kargs)
        return self.cur.fetchall()


    def fetchone(self, query, args=None, log=0, **kargs):
        self.execute(query, args, log=log, **kargs)
        res = self.cur.fetchone()
        if log or self.log_all:
            self.log('=> %s\n' % (res,))
        return res


    def fetchvalue(self, query, args=None, log=0, **kargs):
        res = self.fetchone(query, args=args, log=log, **kargs)
        if res:
            res = res[0]
        return res


    def fetchmany(self, count, query, args=None, log=0, **kargs):
        self.execute(query, args, log=log, **kargs)
        return self.cur.fetchmany(count)


    def get_id(self, val, table, col_val='name', col_key=None, insert=1, case_sens=1, log=0):
        """
        Get the ID (key) of a value (assuming key/value as (row)id/name).
        """
        if type(val) is not tuple:
            val = (val,)
        if col_key is None:
            col_key = self.DEF_ID_KEY
        op = '=' if case_sens else 'LIKE'
        ckey = '%s%s%s' % (col_val, col_key, val)
        if table in self.cacheable and table in self.cache and ckey in self.cache[table]:
            return self.cache[table][ckey]
        sel_col_val = ' AND '.join('%s %s ?' % (x.strip(), op,) for x in col_val.split(','))
        res = self.fetchone("SELECT %s FROM %s WHERE %s" % (col_key, table, sel_col_val), val, log=log)
        if (not res) or res[0] is None:
            if insert:
                ins_col_val = ','.join('?' * len(val))
                self.execute("INSERT INTO %s(%s) VALUES(%s)" % (table, col_val, ins_col_val), val, log=log)
                self.commit()
                res = self.fetchvalue("SELECT %s FROM %s WHERE %s" % (col_key, table, sel_col_val), val, log=log)
            else:
                res = None
        else:
            res = res[0]
        if res is not None and table in self.cacheable:
            if table not in self.cache:
                self.cache[table] = {}
            self.cache[table][ckey] = res
        return res


    def get_value(self, key, table, col_val='name', col_key=None, log=0):
        """
        Get the value for the given ID (key).
        """
        if type(key) is not tuple:
            key = (key,)
        ckey = '%s%s%s' % (key, col_val, col_key)
        if col_key is None:
            col_key = self.DEF_ID_KEY
        if col_key == self.DEF_ID_KEY and table in self.cacheable and table in self.cache and ckey in self.cache[table]:
            return self.cache[table][ckey]
        col_key = ' AND '.join('%s %s' % (x.strip(), '= ?' if key[i] is not None else 'IS NULL')
                for i, x in enumerate(col_key.split(',')))
        res = self.fetchone("SELECT %s FROM %s WHERE %s" % (col_val, table, col_key), [x for x in key if x is not None],
                log=log)
        if res and len(res) == 1:
            res = res[0]
        if log:
            print(' => %s' % res)
        if table in self.cacheable:
            if table not in self.cache:
                self.cache[table] = {}
            self.cache[table][ckey] = res
        return res


    def insert(self, table, fields, keyColumns=('name',), update=1, no_id_key=False, log=0):
        if type(keyColumns) is str:
            keyColumns = (keyColumns,)
        where = ' AND '.join('%s = ?' % (x,) for x in keyColumns)
        keys = tuple([fields[x] for x in keyColumns])
        retkey = ', '.join(keyColumns) if no_id_key else self.DEF_ID_KEY
        rid = self.fetchvalue("SELECT %s FROM %s WHERE %s" % (retkey, table, where), keys, log=log)
        if rid is not None:
            # self.log(' * Known %s: %s; update? :-?' % (keyColumns, keys))
            if update:
                tfields = ', '.join("%s = ?" % (k,) for k in fields)
                self.execute("UPDATE %s SET %s WHERE %s" % (table, tfields, where),
                        tuple(fields.values()) + keys)
        else:
            self.execute("INSERT INTO %s (%s) VALUES (%s)" % (table,
                    ', '.join(fields), ', '.join('?' for x in fields)), fields.values(), log=log)
            rid = self.fetchvalue("SELECT %s FROM %s WHERE %s" % (retkey, table, where), keys, log=log)
        self.conn.commit()
        return rid


    def table_exists(self, table):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def constraint_exists(self, constraint):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def get_table_columns(self, table):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def get_table_sql(self, table):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def dump_table_heads(self, tables, entries=20):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def create_tables(self, tables):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def create_indexes(self, tables):
        raise NotImplementedError("%s() not implemented!" % inspect.stack()[0][3])


    def create_pks(self, pks):
        for table, keys in pks.items():
            name = 'pk_%s' % (table,)
            if not self.constraint_exists(name):
                self.log('Creating PK %s...\n' % name)
                self.execute("ALTER TABLE %s ADD CONSTRAINT %s PRIMARY KEY (%s)" % (table, name, ', '.join(keys)),
                        log=1)


    def dbg(self, msg):
        """
        Print debug message.
        """
        self.log(msg, 'dbg')


    def log(self, msg, mt='info'):
        """
        Print message.
        """
        if self.debug or (mt != 'dbg'):
            sys.stdout.write(msg)


    def dump_qtimes(self, maxlen=100):
        self.log('%-*s  %-6s  %s\n' % (maxlen, 'Query', 'Count', 'Time (s)'))
        for k, v in sorted(self.qtimes.items(), key=lambda x:x[1][1], reverse=1):
            self.log('%-*s  %-6d  %.2f\n' % (maxlen, k[:maxlen], v[0], v[1]))



def caller(fun=None):
    depth = 1 if fun else 2
    return inspect.stack()[depth][3]




#!/usr/bin/env python3
"""
iogp.data.asn1: Parsers for ASN.1 / DER / Windows catalog (.cat) files.

Reference:
    - Certificate ASN.1 grammar:
        - <https://docs.microsoft.com/en-us/windows/win32/seccertenroll/about-x-509-public-key-certificates>

Author: Vlad Topan (vtopan/gmail)
"""
from .oid import OID, OID_MAP

import base64
import re
import sys
import time
import types


ASN1_TAG_BOOLEAN = 0x01
ASN1_TAG_INTEGER = 0x02
ASN1_TAG_BIT_STRING = 0x03
ASN1_TAG_OCTET_STRING = 0x04
ASN1_TAG_NULL = 0x05
ASN1_TAG_OBJECT_IDENTIFIER = 0x06
ASN1_TAG_ENUMERATED = 0x0A
ASN1_TAG_UTF8String = 0x0C
ASN1_TAG_PrintableString = 0x13
ASN1_TAG_TeletexString = 0x14
ASN1_TAG_IA5String = 0x16
ASN1_TAG_UTCTime = 0x17
ASN1_TAG_GeneralizedTime = 0x18
ASN1_TAG_VisibleString = 0x1A
ASN1_TAG_BMPString = 0x1E
ASN1_TAG_SEQUENCE = 0x30
ASN1_TAG_SET = 0x31

ASN1_TAG_NAMES = {v: k[9:] for k, v in globals().items() if k.startswith('ASN1_TAG_')}
ASN1_TAG_VALUES = {v: k for k, v in ASN1_TAG_NAMES.items()}

CRYPT_SUBJ_TYPES = {
    'DE351A42-8E59-11D0-8C47-00C04FC295EE': 'CRYPT_SUBJTYPE_FLAT_IMAGE',
    'C689AABA-8E78-11d0-8C47-00C04FC295EE': 'CRYPT_SUBJTYPE_CABINET_IMAGE',
    'C689AAB8-8E78-11D0-8C47-00C04FC295EE': 'CRYPT_SUBJTYPE_PE_IMAGE',
    'DE351A43-8E59-11D0-8C47-00C04FC295EE': 'CRYPT_SUBJTYPE_CATALOG_IMAGE',
    '9BA61D3F-E73A-11D0-8CD2-00C04FC295EE': 'CRYPT_SUBJTYPE_CTL_IMAGE',
    'C689AAB9-8E78-11D0-8C47-00C04FC295EE': 'CRYPT_SUBJTYPE_JAVACLASS_IMAGE',
    '941C2937-1292-11D1-85BE-00C04FC295EE': 'CRYPT_SUBJTYPE_SS_IMAGE',
    }

KEY_USAGE = [
    'DigitalSignature',
    'NonRepudiation',
    'KeyEncipherment',
    'DataEncipherment',
    'KeyAgreement',
    'KeyCertSign',
    'CRLSign',
    'EncipherOnly',
    'DecipherOnly',
    ]

RX = {
    'utf16le': (rb'^([^\0]\0)+\0\0$',),
    'ascii': (rb'^[\r\n\t\x20-\x7E]+$',),
    }
for k in RX:
    RX[k] = re.compile(*RX[k])



class ASN1Entry:
    """
    ASN.1 entry (has a tag, a value, an offset and a size).
    """

    def __init__(self, data=None, offset=0, decode=True, heuristic=False):
        self.is_decoded = self.is_container = self.is_oid = False
        self.oid_name = None

        if data is not None:
            self.parse(data=data, offset=offset, decode=decode, heuristic=heuristic)

    def parse(self, data=None, offset=None, decode=True, heuristic=False):
        """
        Parse an ASN.1 entry.
        """
        self._data = data
        self.offset = offset
        self._raw_data = None
        b = data[offset + 1]
        multibyte, size, voffset = bool(b & 0x80), b & 0x7F, offset + 2
        if multibyte:
            if size == 0:
                # indefinite form - look for 00 00
                raise NotImplementedError('todo')
            voffset, size = voffset + size, int.from_bytes(data[voffset:voffset + size], byteorder='big')
        # self.raw_data = data[voffset:voffset + size]
        self.tag = data[offset]
        self.data_offset = voffset
        self.data_size = size
        self.size = voffset + size - offset
        self.tag_name = ASN1_TAG_NAMES.get(self.tag)
        if decode:
            self.decode(heuristic=heuristic)

    @property
    def raw_data(self):
        if not self._raw_data:
            self._raw_data = self._data[self.offset:self.offset + self.size]
        return self._raw_data

    def decode(self, heuristic=False):
        """
        Decode a parsed ASN.1 entry.
        """
        tag, offset, size = self.tag, self.data_offset, self.data_size
        raw_data = self.value_raw_data = self._data[offset:offset + size]
        tag_class, tag_value = tag >> 5, tag & 0x1F
        if tag in (ASN1_TAG_SET, ASN1_TAG_SEQUENCE):
            value = ASN1Entry.parse_container(self._data, offset, end=offset + size, decode=True,
                    heuristic=heuristic)
            self.is_container = True
        elif tag == ASN1_TAG_BOOLEAN:
            value = int(bool(self._data[offset]))   # can't subclass bool() to add the .offset attribute
        elif tag in (ASN1_TAG_BIT_STRING, ASN1_TAG_OCTET_STRING):
            value = raw_data
            if tag == ASN1_TAG_BIT_STRING:
                self.ignore_bits = value[0]
                value = value[1:]
            if heuristic and value[0] == 0x30:
                try:
                    value = ASN1Entry.parse_container(self._data,
                            offset + int(tag == ASN1_TAG_BIT_STRING),
                            end=offset + size, decode=True, heuristic=heuristic)
                    self.tag_name = 'Container'
                    self.is_container = True
                except ValueError:
                    pass
        elif tag in (ASN1_TAG_UTCTime, ASN1_TAG_GeneralizedTime):
            value = raw_data.decode('ascii')
        elif tag in (ASN1_TAG_PrintableString, ASN1_TAG_IA5String, ASN1_TAG_TeletexString,
                ASN1_TAG_VisibleString):
            value = raw_data.decode('ascii')
        elif tag == ASN1_TAG_BMPString:
            value = raw_data.decode('UTF-16-BE')
        elif tag == ASN1_TAG_UTF8String:
            value = raw_data.decode('utf8')
        elif tag == ASN1_TAG_NULL:
            value = None
        elif tag == ASN1_TAG_OBJECT_IDENTIFIER:
            value = OID.extract(self._data, offset, size)
            self.is_oid = True
            self.oid_name = OID_MAP.get(value, None)
        elif tag in (ASN1_TAG_INTEGER, ASN1_TAG_ENUMERATED):
            value = int.from_bytes(raw_data, byteorder='big')
        elif tag_class == 0x04: # context-defined
            self.tag_name = 'ContextDefined'
            value = raw_data if size else None
        elif tag_class == 0x05: # context-defined | constructed
            self.tag_name = 'Container'
            value = ASN1Entry.parse_container(self._data, offset, end=offset + size, decode=True,
                    heuristic=heuristic)
            self.is_container = True
        else:
            raise ValueError(f'Unknown type tag: 0x{tag:x} ({tag})!')
        self.is_decoded = True
        if type(value) in (int, str, bytes):
            value = wrap_with_offset_size(value, self.data_offset, self.data_size)
        self.value = value

    @classmethod
    def parse_container(cls, data, offset=0, end=None, max_count=None, decode=True, heuristic=False):
        """
        Parse any number of ASN.1-encoded entries and return a list of raw results.

        :return: List of ASN1Entry().
        """
        result = []
        if end is None:
            end = len(data)
        count = 0
        base_offset = offset
        while offset + 1 < end and ((not max_count) or count < max_count):
            entry = cls(data, offset, decode=decode, heuristic=heuristic)
            result.append(entry)
            offset = entry.data_offset + entry.data_size
            count += 1
        if (not max_count) and offset != end:
            raise ValueError(f'Invalid ASN.1 stream end (last offset: {offset}, data ends at {end})!')
        result = wrap_with_offset_size(result, base_offset, offset - base_offset)
        return result

    def dump(self, title=None, indent=0):
        """
        Dump ASN.1 object.
        """
        res = []
        if title:
            res.append(title)
        value = self.value
        if not self.is_container:
            res.append(f"{indent * '  '}{ASN1_TAG_NAMES.get(self.tag, hex(self.tag))}: {self.str_value}")
        else:
            res.append(f"{indent * '  '}#<{indent}> [{len(value)}]")
            for e in value:
                res.append(e.dump(indent=indent + 1))
        return '\n'.join(res)

    @property
    def str_value(self):
        """
        Render this entry's value a (flattened) string.
        """
        if self.is_container:
            res = '[%s]' % ';'.join(x.str_value for x in self.value)
        elif not isinstance(self.value, bytes):
            res = str(self.oid_name or self.value)
        elif RX['ascii'].search(self.value):
            res = self.value.decode('ascii')
        else:
            res = bytes.hex(self.value)
        res = wrap_with_offset_size(res, self.data_offset, self.data_size)
        return res

    def __getitem__(self, key):
        return self.value[key]

    def __len__(self):
        if self.is_container:
            return len(self.value)
        raise ValueError(f'{self} has no len() (is not a container)!')

    def __repr__(self):
        return f'ASN1Entry<0x{self.tag:02X}:{self.value.__class__.__name__}@{self.offset}[{self.size}]>'

    def is_certificate(self):
        """
        Check if this entry is a certificate.
        """
        return self.is_container and len(self) == 3 and len(self[0]) in (7, 8)

    def identify(self):
        """
        Identify the ASN.1.
        """
        # todo
        '''
        try:
            if len(self) == 3 and len(self[0]) == 6 and self[0][1][0].value == '1.2.840.113549.1.1.5' \
                    and :
                return
        except IndexError:
            return None
        '''



def parse_oid_map(root):
    """
    Parse a list of OID-value entries.
    """
    res = {}
    for e in root.value:
        res[e[0][0].oid_name] = e[0][1].value
    wrap_with_offset_size(res, root.data_offset, root.data_size)
    return res


def wrap_with_offset_size(obj, offset, size):
    obj = types.new_class('ex_' + obj.__class__.__name__, (type(obj),))(obj)
    obj.offset = offset
    obj.size = size
    return obj


def parse_certificate(filename_or_root, offset=0):
    """
    Parse a certificate.

    :param filename_or_root: Filename (str), data (bytes) or ASN1Entry() instance.
    """
    res = wrap_with_offset_size({}, offset, None)
    try:
        if isinstance(filename_or_root, str):
            data = open(filename_or_root, 'rb').read()
        if isinstance(data, bytes):
            root = ASN1Entry(data)
        else:
            root = filename_or_root
        assert len(root) == 3
        info = root[0]
        if len(info) == 7:
            ver = 1
            assert info[0].value == 1
        elif len(info) == 8:
            ver = 2
            assert info[0][0].value == 2
        else:
            raise ValueError(f'Invalid field count: {len(info)}!')
        cert = res['Certificate'] = wrap_with_offset_size({}, info.data_offset, info.data_size)
        cert['Version'] = ver
        assert ver in (1, 2)
        if ver == 1:
            cert['Issued to'] = parse_oid_map(info[1][0][0][0])
            cert['Issued to / hash type'] = info[1][1][1][0].oid_name
            cert['Issued to / hash'] = info[1][1][2].str_value
            cert['Issued by'] = parse_oid_map(info[2][0][0][0])
            cert['Valid from'] = format_asn1_time(info[5][0])
            cert['Valid to'] = format_asn1_time(info[5][1])
        elif ver == 2:
            cert['Serial number'] = hex(info[1].value)[2:]
            cert['Issued by'] = parse_oid_map(info[3])
            cert['Issued to'] = parse_oid_map(info[5])
            cert['Valid from'] = format_asn1_time(info[4][0])
            cert['Valid to'] = format_asn1_time(info[4][1])
            cert['Algorithm'] = info[6][0][0].oid_name
            pkey = ASN1Entry(info[6][1].value)
            pk_offs = info[6][1].data_offset
            pkey[0].data_offset += pk_offs
            pkey[1].data_offset += pk_offs
            cert['Public Key'] = {'Modulus':wrap_with_offset_size(hex(pkey[0].value)[2:],
                    pkey[0].data_offset, pkey[0].data_size), 'Exponent': pkey[1].value}
            cert['Key size'] = int((pkey[0].data_size - 1) * 8)
            cert['Extensions'] = extensions = wrap_with_offset_size({}, info[7][0].data_offset, info[7][0].data_size)
            for e in info[7][0].value:
                if e[0].oid_name == 'Key Usage':
                    entry = ASN1Entry(e[1].value)
                    value = int.from_bytes(entry.value, 'big') >> entry.ignore_bits
                    value = ','.join(KEY_USAGE[i] for i in range(len(KEY_USAGE)) if value & (1 << i))
                    value = wrap_with_offset_size(value, entry.data_offset, entry.data_size)
                elif isinstance(e[1].value, bytes):
                    field = ASN1Entry(e[1].value)
                    if field.is_container:
                        # todo: properly parse the very homogenous ASN.1-encoded additional fields
                        value = field.str_value
                    elif isinstance(field.value, bytes):
                        value = wrap_with_offset_size(bytes.hex(field.value), field.data_offset, field.data_size)
                    else:
                        value = field.value
                else:
                    value = e[1].value
                value.offset += e[1].offset
                extensions[e[0].oid_name] = value
        res['Hash type'] = wrap_with_offset_size(root[1][0].oid_name, root[1][0].data_offset, root[1][0].data_size)
        res['Signature'] = bytes.hex(root[2].value)
    except (IndexError, ValueError, KeyError, AssertionError) as e:
        raise ValueError from e
    return res


def format_asn1_time(entry):
    """
    Parse and format ASN.1 time value.
    """
    value = entry.value
    tvalue = time.strptime(value, f"%{'y' if len(value) == 13 else 'Y'}%m%d%H%M%SZ")
    res = time.strftime("%d.%m.%Y, %H:%M:%S UTC", tvalue) + f' ({value})'
    wrap_with_offset_size(res, entry.data_offset, entry.data_size)
    return res


def parse_catalog(filename_or_root, embedded=None):
    """
    Parse a Windows catalog (.cat) file.

    :param filename_or_root: Filename (str) or ASN1Entry() instance.
    """
    res = {}
    try:
        if isinstance(filename_or_root, str):
            data = open(filename_or_root, 'rb').read()
            root = ASN1Entry(data)
        else:
            root = filename_or_root
        # root.dump()
        # print(root[0])
        assert root.is_container and len(root.value) == 2 and root[0].oid_name == 'signedData'
        certinfo = root[1][0].value
        assert len(certinfo) == 5
        res['Catalog info'] = crtres = {}
        crtres['Version'] = certinfo[0].value
        crtres['Algorithm'] = certinfo[1][0][0].oid_name
        assert not certinfo[1][0][1].value
        assert len(certinfo[2][1]) == 1 and certinfo[2][0].value == '1.3.6.1.4.1.311.10.1' # CTL
        ctl = certinfo[2][1][0].value
        assert len(ctl) in (5, 6) and ctl[0][0].oid_name == 'CATALOG_LIST'
        crtres['Subject usage'] = repr(ctl[0][0].value)
        crtres['List identifier'] = bytes.hex(ctl[1].value)
        crtres['Effective time'] = f'{format_asn1_time(ctl[2])} ({ctl[2].value})'
        assert ctl[3][0].oid_name in ('CATALOG_LIST_MEMBER', 'CATALOG_LIST_MEMBER_V2')
        entries = ctl[4]
        res['Entries'] = crtres = []
        for e in entries:
            crtentry = {}
            crtres.append(crtentry)
            attrs = e[1]
            if RX['utf16le'].search(e[0].value):
                tag = e[0].value.decode('utf16')
            else:
                tag = bytes.hex(e[0].value)
            crtentry['Tag'] = tag.lower()
            for attr in attrs:
                if attr[0].oid_name == 'CAT_MEMBERINFO2_OBJID':
                    assert attr[1][0].value is None
                elif attr[0].oid_name == 'SPC_INDIRECT_DATA_OBJID':
                    if attr[1][0][0][0].oid_name == 'SPC_GLUE_RDN_OBJID':
                        if attr[1][0][0][1][0].value:
                            val = attr[1][0][0][1][0].value
                            try:
                                val = val.decode('UTF-16-BE')
                            except UnicodeError:
                                pass
                            if val != '<<<Obsolete>>>':
                                raise ValueError(f'Unknown SPC_GLUE_RDN_OBJID value: {val}!')
                            crtentry['GLUE-RDN-tag'] = val
                        hashinfo = attr[1][0][1]
                        crtentry['Hash type'] = hashinfo[0][0].oid_name
                        crtentry['Hash'] = bytes.hex(hashinfo[1].value)
                    elif attr[1][0][0][0].oid_name == 'SPC_PE_IMAGE_DATA_OBJID':
                        crtentry['Filetype'] = 'SPC_PE_IMAGE_DATA_OBJID'
                    else:
                        raise ValueError(f'Unknown field type: {attr[1][0][0][0].value}')
                elif attr[0].oid_name == 'CAT_MEMBERINFO_OBJID':
                    clsid = attr[1][0][0].value.strip('{}')
                    subjtype = CRYPT_SUBJ_TYPES[clsid]
                    crtentry['Subject Type'] = subjtype
                    crtentry['Subject CLSID'] = clsid
                elif attr[0].oid_name == 'CAT_NAMEVALUE_OBJID':
                    name, flags, val= attr[1][0].value
                    val = val.value
                    if isinstance(val, bytes):
                        try:
                            val = val.decode('utf16').rstrip('\0')
                        except UnicodeError:
                            pass
                    crtentry[name.value] = f'{val} ({hex(flags.value)})'
                else:
                    raise ValueError(f'Unknown attribute type: {attr[0].oid_name}')
        if len(ctl) > 5:
            assert len(ctl[5]) == 1
            if len(ctl[5][0]):
                res['Name-value entries'] = crtres = {}
            for namevalue in ctl[5][0]:
                assert namevalue[0].oid_name == 'CAT_NAMEVALUE_OBJID'
                nvdata = ASN1Entry.parse_container(namevalue[1].value)
                for e in nvdata:
                    name, flags, v = e.value
                    val = v.value
                    try:
                        val = val.decode('utf16').rstrip('\0')
                    except UnicodeError:
                        pass
                    crtres[name.value] = f'{val} ({hex(flags.value)})'
        res['Certificates'] = crtres = []
        for cert in certinfo[3]:
            crtres.append(parse_certificate(cert))
            if embedded is not None:
                embedded.append(cert.raw_data)
        res['Timestamping'] = crtres = {}
        ts = certinfo[4][0]
        crtres['Hash type'] = ts[2][0].oid_name
        crtres['Algorithm'] = ts[4][0].oid_name
        crtres['Public key'] = bytes.hex(ts[5].value)
        try:
            certs = ts[6][0][1][0][1][0][3].value
            # todo: extract more info about the countersignature
            crtres['Certificates'] = crtlst = []
            for cert in certs:
                if len(cert) == 3:
                    if embedded is not None:
                        embedded.append(cert.raw_data)
                    crtlst.append(parse_certificate(cert))
        except (ValueError, AssertionError, IndexError):
            pass    # todo: fixme - timestamping data needs better parsing
            sys.stderr.write('WARNING: Timestamping countersignature information is incomplete!\n')
    except (IndexError, ValueError, KeyError, AssertionError) as e:
        raise ValueError from e
    return res


def parse_certificate_file(filename):
    """
    Parse .cer or .p7b certificate file.
    """
    data = open(filename, 'rb').read()
    if data.startswith(b'-----BEGIN '):
        data = data.split(b'-----')[2]
        data = base64.b64decode(data)
    return parse_certificate(ASN1Entry(data))



"""
iogp.net: Various network-related tools.

Author: Vlad Topan (vtopan/gmail)
"""
from .net import init, download, load_cookies, USER_AGENTS, GET_HEADERS
from .packet import PCap, Packet, decode_mac, decode_ip, encode_hex, PROTOCOLS


"""
DataView: wrapper over raw binary data, files etc.

Each DataView instance has a unique (numeric) `.id`; if the DataView.open() factory is used,
identical buffers will point to the same DataView when multiple instances are created. The same
filename (case sensitive) will always have the same DataView instance associated with it.

Author: Vlad Topan (vtopan/gmail)
"""
import collections
import hashlib
import mmap
import os
import struct
import threading

from .ds import AttrDict
from .fileid import get_file_type


PREFILTER_SIZE = 2048
MAX_CMP_HASH_SIZE = 64 * 1024 * 1024
BUFFER_ID_LOCK = threading.Lock()


class DataView:
    """
    Data view object: wrapper over raw binary data, files etc.

    :param data: The raw data (bytes or bytearray) or filename.
    :param filename: The file name containing the data.
    :param parent: DataView() instance.
    :param relation: `str`, relation with the parent (if any).
    :param virtual: False if the filename exists (can be overriden).
    """
    MAP = {}                                    # maps `.id`s to DataView() instances
    FILENAME_MAP = {}                           # maps filenames to DataView() instances
    PREFLT_MAP = collections.defaultdict(list)  # maps prefilter-MD5s (on the first part of buffers)
                                                # to lists of DataView() instances


    def __init__(self, data=None, filename=None, name=None, readonly=True, parent=None,
                relation=None, virtual=None):
        if data is None and filename is None:
            data = b''
        self.name = name
        self._full_hash = None
        if type(data) is str:
            filename, data = data, None
        if data is not None and not readonly and type(data) is bytes:
            data = bytearray(data)
        BUFFER_ID_LOCK.acquire()
        self.id = len(DataView.MAP)
        DataView.MAP[self.id] = self
        BUFFER_ID_LOCK.release()
        self.fullname = os.path.abspath(filename) if filename else None
        if virtual is None:
            virtual = not (filename and os.path.isfile(self.fullname))
        if filename and not virtual:
            self.FILENAME_MAP[self.fullname] = self
        self.rawdata = data
        self.filename = filename
        self.readonly = readonly
        self._data = None
        self.mm = self.fh = None
        self.parent = parent
        self.virtual = virtual
        self.relation = relation
        if parent:
            self.parent.children.append(self)
        self.children = []
        self.hashes = None
        self.pf_hash = prefilter_hash(self)
        DataView.PREFLT_MAP[self.pf_hash].append(self)

    @classmethod
    def open(cls, data=None, filename=None, id=None, **kwargs):
        """
        Creates a new DataView or returns an existing one based on the given `data`,
        `filename` or `id`.
        """
        if id:
            return DataView.MAP[id]
        if filename and (not kwargs.get('virtual')) and os.path.abspath(filename) in DataView.MAP:
            return DataView.FILENAME_MAP[os.path.abspath(filename)]
        rdata = data if data else open(filename, 'rb').read(PREFILTER_SIZE)
        pf_hash = prefilter_hash(rdata)
        f_hash = None
        if pf_hash in DataView.PREFLT_MAP:
            for e in DataView.PREFLT_MAP[pf_hash]:
                if f_hash is None:
                    f_hash = full_hash(data) if data else full_hash_from_file(filename)
                if f_hash == e.full_hash:
                    return e
        # new buffer, create DataView() instance
        return DataView(data=data, filename=filename, **kwargs)

    @property
    def full_hash(self):
        """
        "Full" (up to MAX_CMP_HASH_SIZE) file hash.
        """
        if not self._full_hash:
            self._full_hash = full_hash(self.raw)
        return self._full_hash

    @property
    def raw(self):
        """
        This is the actual data. The file view is created automagically as needed and can be closed
        with .close().
        """
        if self._data is not None:
            return self._data
        if self.rawdata is not None:
            self._data = self.rawdata
        else:
            if self.virtual:
                raise ValueError('Attempted to open a virtual filename!')
            mode, access = 'rb', mmap.ACCESS_READ
            if not self.readonly:
                mode = 'rb+'
                access |= mmap.ACCESS_WRITE
            self.fh = open(self.filename, mode)
            self.mm = mmap.mmap(self.fh.fileno(), 0, access=access)
            self._data = self.mm
        return self._data

    def startswith(self, data):
        return self.raw[:len(data)] == data

    @property
    def file_type(self):
        if not hasattr(self, '_file_type'):
            self._file_type = get_file_type(self.raw)
        return self._file_type

    def close(self):
        """
        Close the file (if open).
        """
        if self._data is not None:
            if self.mm:
                self.mm.close()
            if self.fh:
                self.fh.close()
            self.mm = self.fh = None
            self._data = None

    def find(self, pattern, offset=0):
        """
        Find a bytes pattern in the data.
        """
        return self.raw.find(pattern, offset)

    def __setitem__(self, item, value):
        if self.readonly:
            raise ValueError('Readonly flag set!')
        self.raw[item] = value

    def __getitem__(self, item):
        return self.raw[item]

    def __len__(self):
        return len(self.raw)

    def read_I2(self, offset):
        """
        Read a two-byte unsigned int from offset.
        """
        return struct.unpack('<H', self.raw[offset:offset + 2])[0]

    def read_I4(self, offset):
        """
        Read a four-byte unsigned int from offset.
        """
        return struct.unpack('<I', self.raw[offset:offset + 4])[0]

    def read_I8(self, offset):
        """
        Read an eight-byte unsigned int from offset.
        """
        return struct.unpack('<Q', self.raw[offset:offset + 8])[0]

    def read_SI4(self, offset):
        """
        Read a four-byte signed int from offset.
        """
        return struct.unpack('<i', self.raw[offset:offset + 4])[0]

    def save(self):
        """
        Save changes (only relevant if both data and filename were passed to the constructor).
        """
        if self.filename and self.rawdata:
            open(self.filename, 'wb').write(self.rawdata)

    def _calculate_hashes(self):
        hashes = {k: getattr(hashlib, k)() for k in ('md5', 'sha1', 'sha256')}
        chunk_size = 64 * 4096
        for i in range(0, len(self), chunk_size):
            chunk = self[i:i + chunk_size]
            for h in hashes.values():
                h.update(chunk)
        self.hashes = AttrDict({k: v.hexdigest() for k, v in hashes.items()})

    @property
    def md5(self):
        """
        The buffer's MD5 hash.
        """
        if not self.hashes:
            self._calculate_hashes()
        return self.hashes.md5

    @property
    def sha1(self):
        """
        The buffer's SHA-1 hash.
        """
        if not self.hashes:
            self._calculate_hashes()
        return self.hashes.sha1

    @property
    def sha256(self):
        """
        The buffer's SHA-256 hash.
        """
        if not self.hashes:
            self._calculate_hashes()
        return self.hashes.sha256



def prefilter_hash(data):
    """
    Computes the prefilter hash for buffer comparison by contents.
    """
    return hashlib.md5(data[:PREFILTER_SIZE]).hexdigest()


def full_hash(data, tail=None):
    """
    Computes the full hash for buffer comparison by contents.
    """
    h = hashlib.sha256(data[:MAX_CMP_HASH_SIZE])
    if len(data) > MAX_CMP_HASH_SIZE and not tail:
        tail = data[-PREFILTER_SIZE:]
    if tail:
        h.update(tail)
    return h.hexdigest()


def full_hash_from_file(filename):
    """
    Computes the full file-comparison hash for a filename.
    """
    with open(filename, 'rb') as f:
        if os.path.getsize(filename) > MAX_CMP_HASH_SIZE:
            data = f.read(MAX_CMP_HASH_SIZE)
            f.seek(-PREFILTER_SIZE, 2)
            tail = f.read()
        else:
            data, tail = f.read(), None
        return full_hash(data, tail)



"""
iogp.data.ds: Data structures.

Author: Vlad  Topan (vtopan/gmail)
"""
import ast
import copy
import os
import pprint
import re


class AttrDict(dict):
    """
    Keys-as-attributes dictionary. Can be used for storing configuration data.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def save(self, filename, skip_keys=None):
        """
        Save the contents to a file (assuming the instance only contains basic types).

        :param skip_keys: Skip these (top-level) keys when saving (list/tuple).
        """
        data = self
        if skip_keys:
            data = copy.deepcopy(data)
            for k in skip_keys:
                data.pop(k)
        open(filename, 'w', encoding='utf8').write(pprint.pformat(data))

    def load(self, filename, errors='replace', keep_old_keys=False):
        """
        Load the dict from a file (converting top-level dicts to AttrDict).
        """
        if not os.path.isfile(filename):
            raise OSError(f'File {filename} not found!')
        try:
            data = open(filename, encoding='utf8', errors=errors).read()
            if not data.strip():
                # empty file
                return
            data = ast.literal_eval(data)
        except Exception as e:
            raise ValueError(f'Failed parsing file: {e}!') from e
        dict_to_AttrDict(data, self, keep_old_keys=keep_old_keys)

    def eval_path(self, path, path_key='paths'):
        """
        Interpolate a path using the values in self[path_key].
        """
        path = re.sub(r'\$(\w+)', lambda m: self[path_key].get(m.groups()[0], m.groups()[0]), path)
        return os.path.normpath(path)

    def from_dict(self, d):
        """
        Populate from a dict (returning self).
        """
        dict_to_AttrDict(d, self)
        return self



class Config(AttrDict):
    """
    File-based AttrDict for use as a configuration file.
    """

    def __init__(self, filename, template=None):
        super().__init__()
        self._paths = {}
        if template:
            self.from_dict(template)
            self.resolve_cfg_paths()
        self._filename = filename

    def load(self, if_exists=True):
        if if_exists and not os.path.isfile(self._filename):
            return
        super().load(self._filename, keep_old_keys=True)

    def save(self, skip=None):
        """
        Save to file.

        :param skip: Keys to skip.
        """
        skip_keys = ('_filename', '_paths') + tuple(skip or [])
        super().save(self._filename, skip_keys=skip_keys)

    def resolve_cfg_paths(self, paths=None):
        """
        Resolve an AttrDict containing inter-referencing paths and normalize backslash to slash.

        Sample input: `AttrDict(root='/path', subpath='<root>/custompath')`
        """
        paths = paths or self.get('paths', {})
        for k, v in paths.items():
            paths[k] = paths[k].replace('\\', '/')
        any = 1
        paths = AttrDict(paths)     # work on a copy
        while any:
            any = 0
            for k, v in paths.items():
                if m := re.search(r'\$(\w+)', v):
                    any = 1
                    pattern, name = m.group(0, 1)
                    if name not in paths:
                        raise ValueError(f'Invalid path key referenced "{name}" by "{k}"!')
                    paths[k] = v.replace(pattern, paths[name])
        self._paths = paths
        return paths

    def path(self, name):
        """
        Get a resolved path.
        """
        if name not in self._paths and name in self.paths:
            self.resolve_cfg_paths()
        return self._paths[name]



def dict_to_AttrDict(source, dest=None, keep_old_keys=False):
    """
    Recursively convert a dict to an AttrDict.

    :param keep_old_keys: If a dict exists in both source and dest, keep the old keys if no
        corresponding new keys exist in source.
    """
    res = AttrDict() if dest is None else dest
    for k, v in source.items():
        if type(v) is dict:
            v = dict_to_AttrDict(v)
        if k in res and keep_old_keys and isinstance(res[k], dict):
            res[k].update(v)
        else:
            res[k] = v
    return res



#!/usr/bin/env python3
"""
iogp unit tests.

Author: vtopan / gmail
"""
import iogp
import os
import time
import argparse


argp = argparse.ArgumentParser(description='iogp tester')
argp.add_argument('-N', '--network', help='include network-dependent tests', action='store_true') 
args = argp.parse_args() 

test_zip_data = open('test.zip', 'rb').read()


# iogp.archive

a = iogp.data.Archive('test.zip')
assert ['test'] == list(a.list_members())
assert b'test' == a.read('test')
a.close()

# iogp.carve

delta = 10
emb = list(iogp.data.extract_embedded_files(b'X' * delta + test_zip_data + b'X' * delta))
assert len(emb) == 1
assert emb[0][0] == delta
assert emb[0][1] == len(test_zip_data)

# iogp.ds

assert iogp.data.dict_to_AttrDict({'test': []}).test == []

# iogp.fileid

assert iogp.data.get_file_type(test_zip_data) == 'zip'

if os.name == 'nt':
    assert iogp.data.get_file_type(r'c:\windows\regedit.exe') == 'pe+'  # expect Win 64
elif os.name == 'posix':
    assert iogp.data.get_file_type('/bin/sh') == 'elf'

# iogp.str

assert iogp.data.ellipsis('0123456789', 9) == '0123[...]'

# iogp.db

cache = iogp.db.DataCache(max_age=1)
cache['test'] = 'testdata'
assert cache['test'] == 'testdata'
time.sleep(1)
assert cache.get_age('test') > 0.5
assert cache['test'] is None

# iogp.net

assert b'Cloudflare' in iogp.net.download('https://1.1.1.1/')

# iogp.win

if os.name == 'nt':
    assert iogp.win.get_window_size() > (100, 100)

print('All unit tests were successful!')



"""
iogp.db: DB-related tools.

Author: Vlad Topan (vtopan/gmail)
"""
from .sqlite import DB as SqliteDB
from .cache import ObjCache, DataCache, DataStore, encode_data, decode_data, ENC_MODE_ZLIB, ENC_MODE_PICKLE, \
    ENC_MODE_MARSHAL, ENC_MODE_YAML, ENC_MODE_JSON, ENC_MODE_PZ, ENC_MODE_MARZ, ENC_MODE_YZ, ENC_MODE_JZ, \
    CACHE_ALWAYS, CACHE_ONLY, BE_SQLITE, BE_DBM, DEF_MAX_AGE, DEF_DB_FILENAME




"""
iogp.win: ctypes-based wrappers over useful Windows APIs.

Example usage: iterating all windows

    import iogp.win as win
    def callback(hwnd, lparam):
        size = 512
        buf = win.unicode_buffer(size)
        win.GetWindowText(hwnd, buf, size)
        print(f'[-] Handle: {hwnd}, title: {buf.value}')
        return 1  # 0 to stop
    win.EnumWindows(win.WNDENUMPROC(callback), 0)

Author: Vlad Ioan Topan (vtopan/gmail).
"""

import ctypes
from ctypes import (byref, cast, windll, get_last_error, FormatError, c_void_p, c_int,
    c_int64, Structure, POINTER, WINFUNCTYPE, GetLastError, WinError)
from ctypes.wintypes import (BOOL, BYTE, WORD, DWORD, HANDLE, HMODULE, HINSTANCE, HWND, LPWSTR,
        WPARAM, LPARAM, HDC, RECT, POINT, UINT, LPCSTR, LPCWSTR)


user32 = windll.user32
kernel32 = windll.kernel32
gdi32 = windll.gdi32

INVALID_HANDLE_VALUE = c_void_p(-1).value
LRESULT = c_int64
PDWORD = POINTER(DWORD)
PROCESS_QUERY_LIMITED_INFORMATION = 0x1000
PVOID = c_void_p
NULL = None

DIB_RGB_COLORS = 0
ICON_BIG = 1
ICON_SMALL = 0
IMAGE_ICON = 1
LR_CREATEDIBSECTION = 0x20001
LR_DEFAULTSIZE = 0x40
LR_LOADFROMFILE = 0x10
MAX_PATH = 260
SRCCOPY = 0xCC0020
WM_SETICON = 0x0080


# Structures

class RGBQUAD(Structure):
    _fields_ = [
        ("rgbBlue", BYTE),
        ("rgbGreen", BYTE),
        ("rgbRed", BYTE),
        ("rgbReserved", BYTE),
    ]
PRGBQUAD = POINTER(RGBQUAD)     # NOQA


class BITMAPINFOHEADER(Structure):
    _fields_ = [
        ("biSize", DWORD),
        ("biWidth", DWORD),
        ("biHeight", DWORD),
        ("biPlanes", WORD),
        ("biBitCount", WORD),
        ("biCompression", DWORD),
        ("biSizeImage", DWORD),
        ("biXPelsPerMeter", DWORD),
        ("biYPelsPerMeter", DWORD),
        ("biClrUsed", DWORD),
        ("biClrImportant", DWORD),
    ]
PBITMAPINFOHEADER = POINTER(BITMAPINFOHEADER)   # NOQA


class BITMAPINFO(Structure):
    _fields_ = [
        ("bmiHeader", BITMAPINFOHEADER),
        ("bmiColors", RGBQUAD * 1),
    ]
PBITMAPINFO = POINTER(BITMAPINFO)   # NOQA


### Windows APIs

## kernel32

# GetModuleHandle(lpModuleName)
GetModuleHandle = WINFUNCTYPE(HMODULE, LPCWSTR)(("GetModuleHandleW", kernel32))
# OpenProcess(hProcess, dwFlags, lpExeName, lpdwSize)
OpenProcess = WINFUNCTYPE(HANDLE, DWORD, BOOL, DWORD)(("OpenProcess", kernel32)) 
# QueryFullProcessImageName(hObject)
QueryFullProcessImageName = WINFUNCTYPE(BOOL, HANDLE, DWORD, LPWSTR,
        PDWORD)(("QueryFullProcessImageNameW", kernel32))
# CloseHandle(hWnd)
CloseHandle = WINFUNCTYPE(BOOL, HANDLE)(("CloseHandle", kernel32))

## user32

# GetActiveWindow()
GetActiveWindow = WINFUNCTYPE(HWND)(("GetActiveWindow", user32))
# SetActiveWindow(hWnd)
SetActiveWindow = WINFUNCTYPE(HWND, HWND)(("SetActiveWindow", user32))
GetForegroundWindow = WINFUNCTYPE(HWND)(("GetForegroundWindow", user32))
# SetForegroundWindow(hWnd)
SetForegroundWindow = WINFUNCTYPE(BOOL, HWND)(("SetForegroundWindow", user32))
# SetFocus(hWnd)
SetFocus = WINFUNCTYPE(HWND, HWND)(("SetFocus", user32))
# FindWindow(lpClassName, lpWindowName)
FindWindow = WINFUNCTYPE(HWND, LPWSTR, LPWSTR)(("FindWindowW", user32))
# WNDENUMPROC(hwnd, lParam) - callback for EnumWindows
WNDENUMPROC = WINFUNCTYPE(BOOL, HWND, LPARAM) # (lpEnumFunc, lParam)
# EnumWindows(hWnd, lpString, nMaxCount)
EnumWindows = WINFUNCTYPE(BOOL, WNDENUMPROC, LPARAM)(("EnumWindows", user32))
# GetWindowText(hWnd, lpdwProcessId)
GetWindowText = WINFUNCTYPE(c_int, HWND, LPWSTR, c_int)(("GetWindowTextW", user32))
# GetWindowThreadProcessId(dwDesiredAccess, bInheritHandle, dwProcessId)
GetWindowThreadProcessId = WINFUNCTYPE(DWORD, HWND, PDWORD)(("GetWindowThreadProcessId", user32))
GetWindowDC = WINFUNCTYPE(HDC, HWND)(("GetWindowDC", user32))
# GetWindowRect(hWnd, lpRect)
GetWindowRect = WINFUNCTYPE(BOOL, HWND, POINTER(RECT))(("GetWindowRect", user32))
# GetClientRect(hWnd, lpRect)
GetClientRect = WINFUNCTYPE(BOOL, HWND, POINTER(RECT))(("GetClientRect", user32))
ReleaseDC = WINFUNCTYPE(c_int, HWND, HDC)(("ReleaseDC", user32))
# ClientToScreen(hWnd, lpPoint)
ClientToScreen = WINFUNCTYPE(BOOL, HWND, POINTER(POINT))(("ClientToScreen", user32))
# PrintWindow(hwnd, hdcBlt, nFlags)
PrintWindow = WINFUNCTYPE(BOOL, HWND, HDC, UINT)(("PrintWindow", user32))
# FlashWindow(hWnd, bInvert)
FlashWindow = WINFUNCTYPE(BOOL, HWND, BOOL)(("FlashWindow", user32))

## gdi32

# BitBlt(hdc, cx, cy)
BitBlt = WINFUNCTYPE(BOOL, HDC, c_int, c_int, c_int, c_int, HDC, c_int, c_int, DWORD)(("BitBlt", gdi32))
# CreateCompatibleDC(hdc)
CreateCompatibleDC = WINFUNCTYPE(HDC, HDC)(("CreateCompatibleDC", gdi32))
# GetDIBits(hdc, hbm, start, cLines, lpvBits, lpbmi, usage)
GetDIBits = WINFUNCTYPE(c_int, HDC, HANDLE, UINT, UINT, PVOID, PBITMAPINFO, UINT)(("GetDIBits", gdi32))
# SelectObject(hdc, h)
SelectObject = WINFUNCTYPE(HANDLE, HDC, HANDLE)(("SelectObject", gdi32))
# DeleteDC(hDC)
DeleteDC = WINFUNCTYPE(BOOL, HDC)(("DeleteDC", gdi32))
# DeleteObject(ho)
DeleteObject = WINFUNCTYPE(BOOL, HANDLE)(("DeleteObject", gdi32))
# LoadImageW(hInst, name, type, cx, cy, fuLoad)
LoadImage = WINFUNCTYPE(HANDLE, HINSTANCE, LPCWSTR, UINT, c_int, c_int, UINT)(("LoadImageW", user32))
# SendMessage(hWnd, Msg, wParam, lParam)
SendMessage = WINFUNCTYPE(LRESULT, HWND, UINT, WPARAM, LPARAM)(("SendMessageA", user32))


### Useful wrappers

def last_error():
    err = get_last_error()
    return "0x%08X: %s" % (err, FormatError(err))


def unicode_buffer(size):
    """
    Create a WSTR buffer of `size` characters.
    """
    return ctypes.create_unicode_buffer(size)


def get_window_rect(title=None):
    """
    Get the coordinates of the window given by title (default: the top window).

    :return: tuple(x1, y1, x2, y2) or None (if window not found)
    """
    if title:
        hwnd = FindWindow(None, title)
        if not hwnd:
            return None
    else:
        hwnd = GetForegroundWindow()
    rect = RECT()
    GetWindowRect(hwnd, byref(rect))
    return (rect.left, rect.top, rect.right, rect.bottom)


def get_window_size(title=None):
    """
    Get the coordinates of the window given by title (default: the top window).

    :return: tuple(width, height) or None (if window not found)
    """
    if title:
        hwnd = FindWindow(None, title)
        if not hwnd:
            return None
    else:
        hwnd = GetForegroundWindow()
    rect = RECT()
    GetClientRect(hwnd, byref(rect))
    return (rect.right, rect.bottom)


def screenshot(hwnd=None, include_titlebar=True, as_rgb=True):
    """
    Take a screenshot of the given window (if None, use the current active window).

    :param hwnd: Window handle (or None).
    :param include_titlebar: Include the window's title bar in the screenshot.
    :param as_rgb: Return the raw bytes as 'RGB' instead of 'BGRX'.
    :return: tuple(width, height, raw_bmp_data)
    """
    if hwnd is None:
        hwnd = GetForegroundWindow()
    dc = GetWindowDC(hwnd)
    if not hwnd:
        raise OSError(f'GetForegroundWindow failed: {last_error()}')
    cdc = CreateCompatibleDC(dc)
    if not cdc:
        raise OSError(f'CreateCompatibleDC failed: {last_error()}')
    rect = RECT()
    fun = GetWindowRect if include_titlebar else GetClientRect
    if not fun(hwnd, byref(rect)):
        raise OSError(f'GetClient/WindowRect failed: {last_error()}')
    left, right, top, bottom = rect.left, rect.right, rect.top, rect.bottom
    width = right - left
    height = bottom - top
    bmp = CreateCompatibleBitmap(dc, width, height)
    if not bmp:
        raise OSError(f'CreateCompatibleBitmap failed: {last_error()}')
    if not SelectObject(cdc, bmp):
        raise OSError(f'SelectObject failed: {last_error()}')
    if not PrintWindow(hwnd, cdc, 0 if include_titlebar else 1):
        raise OSError(f'PrintWindow failed: {last_error()}')
    bmph = BITMAPINFOHEADER(ctypes.sizeof(BITMAPINFOHEADER), width, height, 1, 32, 0)
    buffer = ctypes.create_string_buffer(width * height * 4)
    cnt = GetDIBits(cdc, bmp, 0, height, buffer, cast(byref(bmph), PBITMAPINFO), DIB_RGB_COLORS)
    if not cnt:
        raise OSError(f'GetDIBits failed: {last_error()}')
    DeleteDC(cdc)
    ReleaseDC(hwnd, dc)
    DeleteObject(bmp)
    if as_rgb:
        rgb = ctypes.create_string_buffer(width * height * 3)
        rgb[0::3] = buffer[2::4]
        rgb[1::3] = buffer[1::4]
        rgb[2::3] = buffer[0::4]
        buffer = rgb
    return (width, height, buffer)


def set_window_icon(icon_filename, hwnd=None):
    """
    Set the icon for a window (default: foreground window).
    """
    h_icon = LoadImage(None, icon_filename, IMAGE_ICON, 0, 0, LR_LOADFROMFILE | LR_DEFAULTSIZE)
    if not h_icon:
        raise OSError(f'Failed loading icon from file {icon_filename}: {WinError(GetLastError())}')
    if hwnd is None:
        hwnd = GetForegroundWindow()
    SendMessage(hwnd, WM_SETICON, ICON_BIG, h_icon)
    SendMessage(hwnd, WM_SETICON, ICON_SMALL, h_icon)



"""
iogp.net.net: Basic network-related operations.

Todo: refactor (particularly the error / exception handling).

Author: Vlad Topan (vtopan/gmail)
"""

from ..db import DataCache, CACHE_ONLY, CACHE_ALWAYS, DEF_MAX_AGE     # NOQA

import re
import gzip
import hashlib
import io
import ssl
from urllib.request import HTTPPasswordMgrWithDefaultRealm, HTTPCookieProcessor, HTTPBasicAuthHandler, build_opener, \
    Request, HTTPSHandler
from urllib.parse import parse_qs
from http.cookiejar import LWPCookieJar


USER_AGENTS = {
    'iogp': 'Mozilla/5.0 (compatible; iogp 1.0)',
    # contemporary browsers @ 2018.10
    'chrome69-win10': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
    'chrome69-win7': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
    'chrome69-osx10': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
    'chrome69-linux': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
    'firefox40-win7':'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0) Gecko/20100101 Firefox/40.1',
    'firefox62-win10': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:62.0) Gecko/20100101 Firefox/62.0',
    'firefox62-win81': 'Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:62.0) Gecko/20100101 Firefox/62.0',
    'firefox62-ubuntu': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:62.0) Gecko/20100101 Firefox/62.0',
    'safari12-osx10': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Safari/605.1.15',
    'edge17-win10': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.140 Safari/537.36 Edge/17.17134',
    # old browsers
    'firefox13-win7': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:13.0) Gecko/20100101 Firefox/13.0.1',
    'ie6-winxp':'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1)',
    'ie7-winxp':'Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; SV1)',
    'ie8-winxp':'Mozilla/4.0 (Compatible; MSIE 8.0; Windows NT 5.2; Trident/6.0)',
    'ie9-win7':'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)',
    'ie9-winvista':'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/4.0; GTB7.4; InfoPath.3; SV1; .NET CLR 3.1.76908; WOW64; en-US)',
    'ie10-win7':'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)',
    'opera12-win7':'Opera/9.80 (Windows NT 6.1; U; es-ES) Presto/2.9.181 Version/12.00',
    # bots
    'google-bot': 'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)',
    'bing-bot': 'Mozilla/5.0 (compatible; Bingbot/2.0; +http://www.bing.com/bingbot.htm)',
    'slurp-bot': 'Mozilla/5.0 (compatible; Yahoo! Slurp; http://help.yahoo.com/help/us/ysearch/slurp)',
    'duck-bot': 'DuckDuckBot/1.0; (+http://duckduckgo.com/duckduckbot.html)',
    'baidu-bot': 'Mozilla/5.0 (compatible; Baiduspider/2.0; +http://www.baidu.com/search/spider.html)',
    'yandex-bot': 'Mozilla/5.0 (compatible; YandexBot/3.0; +http://yandex.com/bots)',
    'facebook-bot': 'facebookexternalhit/1.1 (+http://www.facebook.com/externalhit_uatext.php)',
    }
USER_AGENTS[None] = USER_AGENTS['iogp']      # default

# used for a more faitful emulation of a browser's requests
GET_HEADERS = {
    'ff': '''GET /%(path)s HTTP/1.1
Host: %(domain)s
User-Agent: Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.5) Gecko/20070713 Firefox/2.0.0.5
Accept: text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5
Accept-Language: en-us,en;q=0.5
Accept-Encoding: gzip,deflate
Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
Keep-Alive: 300
Connection: keep-alive

''',
    'ie':'''GET /%(path)s HTTP/1.1
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/x-shockwave-flash, */*
Accept-Language: en-us
Accept-Encoding: gzip, deflate
User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1) ; .NET CLR 2.0.50727; MEGAUPLOAD 1.0; FDM)
Host: %(domain)s
Connection: Keep-Alive

''',
    'opr':'''GET /%(path)s HTTP/1.1
User-Agent: Opera/9.25 (Windows NT 5.1; U; en)
Host: %(domain)s
Accept: text/html, application/xml;q=0.9, application/xhtml+xml, image/png, image/jpeg, image/gif, image/x-xbitmap, */*;q=0.1
Accept-Language: en-US,en;q=0.9
Accept-Charset: iso-8859-1, utf-8, utf-16, *;q=0.1
Accept-Encoding: deflate, gzip, x-gzip, identity, *;q=0
Connection: Keep-Alive

'''
    }


_RX = {
    'ver':(r'(\d+)\.(\d+)\.(\d+)(?:\.(\d+))?(?:\s+\((\d+/\d+/\d+ \d+:\d+:\d+)|(\d+:\d+:\d+, \d+\.\d+\.\d+)\))?\s*(.*?)',),
    'user:pass@domain':(r'([^:]+)[:]([^@]+)@([^:]+)(?::(.+))?', re.I),
    'domain':(r'(?:[^/]+\.)*([^/]+\.[a-zA-Z]{2,5})', re.I),
    }
for k in _RX:
    _RX[k] = re.compile(*_RX[k])

__ENV_EVAL__ = {
    }

_DOWNLOAD_CACHE = None


def parse_query(query):
    """
    Parse a query string into a dict.
    """
    res = {}
    for k, v in parse_qs(query).items():
        if isinstance(k, bytes):
            k = k.decode('ascii', 'ignore')
        v = v[0]
        if isinstance(v, bytes):
            v = v.decode('ascii', 'ignore')
        res[k] = v
    return res


def load_cookies(filename):
    """
    Load cookies from file.
    """
    return LWPCookieJar(filename)


def download(url, referer=None, cache=DEF_MAX_AGE, headers=None, cookies=None, auth=None, nocert=False,
            post_data=None, timeout=10, user_agent=None, max_size=None, info=None, ignore_404=True,
            debug=0):
    """
    Download from a URL.

    Todo: logging callback.

    :param auth: Tuple (user, password).
    :param cookies: Jar filename which can be loaded with load_cookies().
    :param headers: dict containing additional headers.
    :param max_size: integer or a (callback) function taking an integer which return True to download.
    :param info: dict which, if given, will be filled in with download info.
    :param nocert: Don't verify server SSL certificate.
    :return: Data as `bytes` (pass the `info` dict to get a HTTP 'code' field and response headers under 'headers').
    """
    if info is None:
        info = {}
    info['error'] = None
    # print(f'[GET] {cache} {url}')
    if cache:
        if _DOWNLOAD_CACHE is None:
            init()
        post_hash = hashlib.md5(post_data if type(post_data) == bytes else str(post_data).encode('utf8')).hexdigest() \
            if post_data else ''
        cache_url = url + post_hash
        data = _DOWNLOAD_CACHE.get(cache_url, cache)
        if data:
            # print(f'Cache hit: {url}')
            return data
        else:
            # print(f'Cache miss: {url}')
            if cache == CACHE_ONLY:
                return None
            pass
    else:
        data = None
    # prepare handlers
    handlers = []
    if cookies is not None:
        handlers.append(HTTPCookieProcessor(cookies))
    if auth:
        # fixme
        toplevel = ('/'.join(url.split('/', 4)[:3]) if '://' in url else ('http://' + url.split('/', 1)[0])) + '/'
        pwmgr = HTTPPasswordMgrWithDefaultRealm()
        user, psw = auth
        pwmgr.add_password(None, toplevel, user, psw)
        handlers.append(HTTPBasicAuthHandler(pwmgr))
    if nocert:
        handlers.append(HTTPSHandler(context=ssl._create_unverified_context()))
    # create opener
    opener = build_opener(*handlers)
    # prepare headers
    if user_agent in USER_AGENTS:
        user_agent = USER_AGENTS[user_agent]
    opener.addheaders = [('User-agent', user_agent)]
    if referer:
        opener.addheaders.append(('Referer', referer))
    for k, v in (headers or {}).items():
        opener.addheaders.append((k, v))
    resp = None
    try:
        # if post_data and type(post_data) != bytes:
        #    post_data = urlencode(post_data).encode('ascii')
        req = Request(url, data=post_data)
        if debug:
            print(f' * net.download({url})')
        resp = opener.open(req, timeout=timeout)
        info['code'] = resp.getcode()
        info['headers'] = resp_info = dict(resp.info())
        if max_size and 'Content-Length' in resp_info:
            # try:
                content_size = int(resp_info['Content-Length'])
                if hasattr(max_size, '__call__'):
                    if not max_size(content_size):
                        return None
                elif content_size > max_size:
                    return None
            # except:
                # [todo] fixme
                # pass
        data = resp.read()
    except Exception as e:
        # fixme
        # if not (ignore_404 and ('urlopen error timed out' in str(e) or \
        #        'getaddrinfo failed' in str(e) or '404' in str(e))):
        info['error'] = str(e)
        print('EXCEPTION: %s' % str(e))
        import traceback
        traceback.print_exc(5)
        return None
    if cookies is not None:
        cookies.extract_cookies(resp, req)
    if data.startswith(b'\x1F\x8B\x08\0\0\0\0\0'):
        data = gzip.decompress(data)
    if data and cache:
        _DOWNLOAD_CACHE[cache_url] = data
    return data


def init(filename='.url-cache.db'):
    global _DOWNLOAD_CACHE
    _DOWNLOAD_CACHE = DataCache(filename)




"""
iogp.data.archive: Archive processing APIs.

On Windows, archiveint.dll (libarchive 3.3.2?) is used (ships with Windows 10).
Code from libarchive's archive.h is used, license in external/libarchive.license.

Author: Vlad Topan (vtopan/gmail)
"""
import os
import ctypes
from ctypes import c_void_p, Structure, c_int, c_char_p, POINTER, byref
if os.name == 'nt':
    from ctypes import WINFUNCTYPE, windll


ARCHIVE_FORMAT_BASE_MASK = 0xff0000
ARCHIVE_FORMAT_CPIO = 0x10000
ARCHIVE_FORMAT_CPIO_POSIX = (ARCHIVE_FORMAT_CPIO | 1)
ARCHIVE_FORMAT_CPIO_BIN_LE = (ARCHIVE_FORMAT_CPIO | 2)
ARCHIVE_FORMAT_CPIO_BIN_BE = (ARCHIVE_FORMAT_CPIO | 3)
ARCHIVE_FORMAT_CPIO_SVR4_NOCRC = (ARCHIVE_FORMAT_CPIO | 4)
ARCHIVE_FORMAT_CPIO_SVR4_CRC = (ARCHIVE_FORMAT_CPIO | 5)
ARCHIVE_FORMAT_CPIO_AFIO_LARGE = (ARCHIVE_FORMAT_CPIO | 6)
ARCHIVE_FORMAT_SHAR = 0x20000
ARCHIVE_FORMAT_SHAR_BASE = (ARCHIVE_FORMAT_SHAR | 1)
ARCHIVE_FORMAT_SHAR_DUMP = (ARCHIVE_FORMAT_SHAR | 2)
ARCHIVE_FORMAT_TAR = 0x30000
ARCHIVE_FORMAT_TAR_USTAR = (ARCHIVE_FORMAT_TAR | 1)
ARCHIVE_FORMAT_TAR_PAX_INTERCHANGE = (ARCHIVE_FORMAT_TAR | 2)
ARCHIVE_FORMAT_TAR_PAX_RESTRICTED = (ARCHIVE_FORMAT_TAR | 3)
ARCHIVE_FORMAT_TAR_GNUTAR = (ARCHIVE_FORMAT_TAR | 4)
ARCHIVE_FORMAT_ISO9660 = 0x40000
ARCHIVE_FORMAT_ISO9660_ROCKRIDGE = (ARCHIVE_FORMAT_ISO9660 | 1)
ARCHIVE_FORMAT_ZIP = 0x50000
ARCHIVE_FORMAT_EMPTY = 0x60000
ARCHIVE_FORMAT_AR = 0x70000
ARCHIVE_FORMAT_AR_GNU = (ARCHIVE_FORMAT_AR | 1)
ARCHIVE_FORMAT_AR_BSD = (ARCHIVE_FORMAT_AR | 2)
ARCHIVE_FORMAT_MTREE = 0x80000
ARCHIVE_FORMAT_RAW = 0x90000
ARCHIVE_FORMAT_XAR = 0xA0000
ARCHIVE_FORMAT_LHA = 0xB0000
ARCHIVE_FORMAT_CAB = 0xC0000
ARCHIVE_FORMAT_RAR = 0xD0000
ARCHIVE_FORMAT_7ZIP = 0xE0000
ARCHIVE_FORMAT_WARC = 0xF0000
ARCHIVE_FORMAT_RAR_V5 = 0x100000

ARCHIVE_EOF    = 1            # Found end of archive.
ARCHIVE_OK = 0          # Operation was successful.
ARCHIVE_RETRY = (-10)   # Retry might succeed.
ARCHIVE_WARN = (-20)    # Partial success.
ARCHIVE_FAILED = (-25)  # Current operation cannot complete.
ARCHIVE_FATAL = (-30)   # No more operations are possible.

MODE_READ = 0
MODE_WRITE = 1

DEF_BUFFER_SIZE = 16384

struct_archive = c_void_p
struct_archive_entry = c_void_p

if os.name == 'nt':
    ARCHIVEINT_LIB = windll.archiveint
    # todo: load libarchive.so on POSIX if available instead of crashing

    get_zlib_version = WINFUNCTYPE(c_char_p)(('archive_zlib_version', ARCHIVEINT_LIB))
    get_zlib_version.errcheck = lambda res, _1, _2: res.decode('utf8')

    archive_read_new = WINFUNCTYPE(struct_archive)(('archive_read_new', ARCHIVEINT_LIB))
    archive_read_support_filter_all = WINFUNCTYPE(c_int, struct_archive)(('archive_read_support_filter_all', ARCHIVEINT_LIB))
    archive_read_support_format_all = WINFUNCTYPE(c_int, struct_archive)(('archive_read_support_format_all', ARCHIVEINT_LIB))
    archive_read_support_format_raw = WINFUNCTYPE(c_int, struct_archive)(('archive_read_support_format_raw', ARCHIVEINT_LIB))
    archive_read_support_compression_all = WINFUNCTYPE(c_int, struct_archive)(('archive_read_support_compression_all', ARCHIVEINT_LIB))
    archive_read_open_filename = WINFUNCTYPE(c_int, struct_archive, c_char_p, c_int)(('archive_read_open_filename', ARCHIVEINT_LIB))
    archive_read_open_memory = WINFUNCTYPE(c_int, struct_archive, c_char_p, c_int)(('archive_read_open_memory', ARCHIVEINT_LIB))
    archive_read_next_header = WINFUNCTYPE(c_int, struct_archive, POINTER(struct_archive_entry))(('archive_read_next_header', ARCHIVEINT_LIB))
    archive_entry_pathname = WINFUNCTYPE(c_char_p, struct_archive_entry)(('archive_entry_pathname', ARCHIVEINT_LIB))
    archive_entry_pathname.errcheck = lambda res, _1, _2: res.decode('utf8') if res else ''
    archive_read_data = WINFUNCTYPE(c_int, struct_archive, c_char_p, c_int)(('archive_read_data', ARCHIVEINT_LIB))
    archive_read_data_skip = WINFUNCTYPE(c_int, struct_archive)(('archive_read_data_skip', ARCHIVEINT_LIB))
    archive_read_data_block = WINFUNCTYPE(c_int, struct_archive, POINTER(c_char_p), POINTER(c_int), POINTER(c_int))(('archive_read_data_block', ARCHIVEINT_LIB))
    archive_read_free = WINFUNCTYPE(c_int, struct_archive)(('archive_read_free', ARCHIVEINT_LIB))
    archive_read_close = WINFUNCTYPE(c_int, struct_archive)(('archive_read_close', ARCHIVEINT_LIB))
    archive_compression_name = WINFUNCTYPE(c_char_p, struct_archive_entry)(('archive_compression_name', ARCHIVEINT_LIB))
    archive_compression_name.errcheck = lambda res, _1, _2: res.decode('utf8') if res else ''
    archive_format_name = WINFUNCTYPE(c_char_p, struct_archive_entry)(('archive_format_name', ARCHIVEINT_LIB))
    archive_format_name.errcheck = lambda res, _1, _2: res.decode('utf8') if res else ''
    archive_error_string = WINFUNCTYPE(c_char_p)(('archive_error_string', ARCHIVEINT_LIB))
    archive_error_string.errcheck = lambda res, _1, _2: res.decode('utf8') if res else ''
else:
    # todo: find alternatives on Linux
    archive_read_new = None
    archive_read_support_filter_all = None
    archive_read_support_format_all = None
    archive_read_support_compression_all = None
    archive_read_open_filename = None
    archive_read_open_memory = None
    archive_read_free = None
    archive_read_next_header = None
    archive_entry_pathname = None
    archive_read_data_block = None
    archive_compression_name = None
    archive_format_name = None


class ArchiveError(ValueError): pass



class Archive:
    """
    Archive object.
    """

    def __init__(self, filename_or_data=None):
        self.mode = MODE_READ
        self.archive = None
        if filename_or_data:
            self.open_archive(filename_or_data)


    def last_error(self):
        """
        Returns the last error from libarchive.
        """
        return archive_error_string()


    def open_archive(self, filename_or_data, buffer_size=DEF_BUFFER_SIZE):
        """
        Open an archive file or buffer for reading.
        """
        self.mode = MODE_READ
        self.buffer_size = buffer_size
        self.source = filename_or_data


    def reset(self):
        """
        Reset (reopen) the archive.
        """
        self.close()
        self.archive = arc = archive_read_new()
        if archive_read_support_filter_all(arc) != ARCHIVE_OK:
            raise ArchiveError(f'Failed enabling filters: {self.last_error()})!')
        if archive_read_support_format_all(arc) != ARCHIVE_OK:
            raise ArchiveError(f'Failed enabling formats: {self.last_error()})!')
        #if archive_read_support_compression_all(arc) != ARCHIVE_OK:
        #    raise ArchiveError(f'Failed enabling compression: {self.last_error()})!')
        if type(self.source) is str:
            res = archive_read_open_filename(arc, self.source.encode('utf8'), self.buffer_size)
        else:
            res = archive_read_open_memory(arc, self.source, self.buffer_size)
        if res != ARCHIVE_OK:
            raise ArchiveError(f'Failed opening file or buffer: {res} ({self.last_error()})!')


    def close(self):
        """
        Close the archive.
        """
        if self.archive:
            archive_read_free(self.archive)
        self.archive = None


    def walk(self, callback=None):
        """
        Iterate archive member filenames returning tuples (name, entry_object).
        """
        if (not self.mode == MODE_READ) and self.source:
            raise ArchiveError('No archive is opened for reading!')
        self.reset()
        entry = struct_archive_entry()
        while archive_read_next_header(self.archive, ctypes.byref(entry)) == ARCHIVE_OK:
            entry_name = archive_entry_pathname(entry)
            if callback:
                callback(entry_name, entry)
            yield entry_name, entry



    def list_members(self):
        """
        Iterate over member names.
        """
        for name, e in self.walk():
            yield name


    def read(self, filename=None):
        """
        Read an archive member (the current one if none given); returns None if the filename is given and not present.

        Do NOT use this for large files - the data will be read to memory.
        """
        if not filename:
            return self.read_crt_entry()
        for name, e in self.walk():
            if name == filename:
                return self.read_crt_entry()


    def read_crt_entry(self):
        """
        Read the current archive entry.
        """
        if (not self.mode == MODE_READ) and self.archive:
            raise ArchiveError('No archive is opened for reading!')
        chunk = c_char_p()
        chunk_size = c_int()
        offset = c_int(0)
        lst = []
        while 1:
            res = archive_read_data_block(self.archive, byref(chunk), byref(chunk_size), byref(offset))
            offset.value += chunk_size.value
            if res == ARCHIVE_EOF:
                break
            elif res != ARCHIVE_OK:
                raise ArchiveError(f'Failed reading current archive entry: {res} ({self.last_error()})!')
            lst.append(chunk.value[:chunk_size.value])
        return b''.join(lst)


    def get_format(self):
        """
        Returns the format of the current archive.
        """
        if (not self.mode == MODE_READ) and self.source:
            raise ArchiveError('No archive is opened for reading!')
        if not self.archive:
            self.reset()
        return archive_compression_name(self.archive)


    def get_entry_format(self):
        """
        Returns the format of the current archive entry.
        """
        if (not self.mode == MODE_READ) and self.archive:
            raise ArchiveError('No archive is opened for reading!')
        return archive_format_name(self.archive)



"""
Example:

    a = Archive('test.zip')
    print(a.get_format())
    for name in a.list_members():
        print(name, a.get_entry_format())
        if name == 'extractme.ext':
            data = a.read()
            open(name, 'wb').write(data)
    print(a.read('test.py'))
    a.close()
"""



#!/usr/bin/env python3
"""
iogp.net.packet: Network packet and .pcap processing library.

Author: Vlad Topan (vtopan/gmail)
"""

import binascii
import struct


BASE_TCPDUMP    = 1000
BASE_ETHER      = 2000
BASE_IP         = 3000

# TCPDump LL IDs (1000 + IDs from http://www.tcpdump.org/linktypes.html)
Ethernet        = BASE_TCPDUMP + 1
IP              = BASE_TCPDUMP + 101

# EtherType-based (2000 + EtherType)
IPv4            = BASE_ETHER + 0x0800
IPv6            = BASE_ETHER + 0x86DD
ARP             = BASE_ETHER + 0x0806
CDP             = BASE_ETHER + 0x2000
LLDP            = BASE_ETHER + 0x88CC
IEEE8021Q       = BASE_ETHER + 0x8100

# IP protocols (3000 + IP protocol number)
ICMP            = BASE_IP + 1
IGMP            = BASE_IP + 2
IPv4Inv4        = BASE_IP + 4
TCP             = BASE_IP + 6
UDP             = BASE_IP + 17
IPv6Inv4        = BASE_IP + 41


# Protocol info: (name, layer);
PROTOCOLS = {
    Ethernet:('Ethernet', 1),
    ARP:('ARP', 2),
    CDP:('CDP', 2),
    LLDP:('LLDP', 2),
    IP:('IP', 3),                   # named "RAW", which means IPv4/6
    IPv4:('IPv4', 3),
    IPv6:('IPv6', 3),
    ICMP:('ICMP', 3),
    IGMP:('IGMP', 3),
    UDP:('UDP', 4),
    TCP:('TCP', 4),
    }



class PCap:
    """
    Wraps a packet capture file.

    @param source: The data source (a filename or raw binary data).
    """

    def __init__(self, source=None):
        self.packets = []
        self.data = b''
        if source:
            self.parse(source)


    def parse(self, source):
        """
        Parses a packet capture file.

        :param source: The data source (a filename or raw binary data).
        """
        if type(source) == str:
            data = open(source, 'rb').read()
            self.filename = source
        else:
            data = source
            self.filename = '<rawdata>'
        self.bom = bom = '>' if data[:4] == b'\xA1\xB2\xC3\xD4' else '<'
        hdr = struct.unpack(bom + 'HHlLLL', data[4:24])
        self.ver = (hdr[0], hdr[1])
        self.max_len = hdr[4]
        self.layer = hdr[5]     # link-layer header type
        pos = 24
        num = 1
        while pos < len(data):
            ts_sec, ts_us, size, orig_size = struct.unpack(bom + 'LLLL', data[pos:pos + 16])
            pos += 16
            pdata = data[pos:pos + size]
            pos += size
            if size < orig_size:
                pdata += b'\0' * (orig_size - size)
            packet = Packet(pdata, ts=ts_sec, ts_us=ts_us, proto=BASE_TCPDUMP + self.layer, num=num)
            self.packets.append(packet)
            num += 1
        return self.packets


    def __str__(self):
        return "PCap[%s]:[%d.%d/%s/%d/%d]" % ((self.filename,) + self.ver + (self.bom, self.max_len, self.layer))



class Packet:
    """
    Basic network packet parser.

    :param data: The data buffer (may be longer than a packet).
    :param layer: The layer/protocol of the packet (see PROTOCOLS); default: 1 = Ethernet.
    :param num: Packet number inside PCAP (1-based).
    :param ts: Timestamp (UTC), as int.
    :param ts_us: Timestamp microseconds as int.
    """

    def __init__(self, data, proto=Ethernet, parent=None, num=0, ts=None, ts_us=None):
        self.ts = ts
        self.ts_us = ts_us
        self.num = num
        self.proto = proto
        self.layers = {}
        self.fields = {}
        self.parent = parent
        self.next = self.next_type = None
        self.data = data
        self.hsize = len(data)
        if self.proto in _FIELDS:
            for finfo in _FIELDS[self.proto]:
                fname, fval, pos = finfo[:3]
                if pos == 'raw':
                    pass
                elif pos == 'self':
                    fval = fval(self)
                else:
                    pos = finfo[2]
                    if type(fval) == str:
                        pat = fval
                        nibble = None
                        if fval[0] == 'n':
                            # nibble
                            pat = 'B'
                            nibble = fval[1]
                        fval = struct.unpack('>' + pat, data[pos:pos + struct.calcsize(pat)])[0]
                        if nibble:
                            fval = (fval >> 4) if nibble == 'h' else (fval & 0xF)
                    else:
                        fval = fval(data[pos:pos + finfo[3]])
                setattr(self, fname, fval)
                self.fields[fname] = fval
        if self.proto == Ethernet:
            # Ethernet
            if self.type == IEEE8021Q:
                self.vlan, self.type = struct.unpack('>HH', data[14:18])
                self.hsize += 4
            self.next_type = BASE_ETHER + self.type
        elif self.proto in [IPv4, IPv6]:
            self.next_type = self.protocol + BASE_IP
        if self.next_type:
            self.payload = data[self.hsize:]
            self.next = Packet(self.payload, proto=self.next_type, parent=self)
            self.payload = self.payload[:len(self.next)]
            self.data = data[:self.hsize + len(self.payload)]


    def __len__(self):
        return getattr(self, 'size', len(self.data) if self.data else 0)


    def __str__(self):
        _str = getattr(self, '_str', ' '.join('%02X' % c for c in self.data[:16]) + '...') % self.__dict__
        s = '%s[%d] %s' % (PROTOCOLS.get(self.proto, ('?/%d' % self.proto,))[0], len(self), _str)
        s += '\nRAW:' + encode_hex(self.data[:self.hsize])[:64]
        if self.next:
            s += '\n' + '\n'.join(['  ' + x for x in str(self.next).split('\n')])
        elif len(self) != self.hsize:
            s += '\n' + 'PAYLOAD: ' + encode_hex(self.data[self.hsize:])[:64]
        return s



def decode_mac(x):
    """
    Decode a MAC address from a binary (bytes) string.
    """
    return ':'.join('%02X' % e for e in x)


def decode_ip(x):
    """
    Decode an IP address (v4 or v6) from a binary (bytes) string.
    """
    if type(x) != bytes:
        x = list(ord(e) for e in x)
    if len(x) == 4:
        # ipv4
        return '.'.join(str(e) for e in x)
    else:
        # ipv6
        return ':'.join('%02X' % e for e in x)


def encode_hex(x):
    """
    Encode a binary string as hex.
    """
    return binascii.hexlify(x).upper().decode('utf8')


def unpack_be_gen(itype):
    def _unpack(data):
        return struct.unpack('>%s' % itype, data[:struct.calcsize(itype)])[0]
    _unpack.__name__ = 'unpack_be_%s' % itype
    return _unpack


# need the decoding functions to be declared for this
_FIELDS = {
    Ethernet: [
        ('vlan', None, 'raw'),
        ('dst', decode_mac, 0, 6),
        ('src', decode_mac, 6, 6),
        ('type', 'H', 12),
        ('hsize', 14, 'raw'),
        ('_str', '%(src)s->%(dst)s', 'raw'),
        ],
    IPv4: [
        ('ver', 'nh', 0),
        ('ihl', 'nl', 0),
        ('size', 'H', 2),
        ('hsize', lambda x:x.ihl * 4, 'self'),
        ('protocol', 'B', 9),
        ('src', decode_ip, 12, 4),
        ('dst', decode_ip, 16, 4),
        ('_str', '%(src)s->%(dst)s', 'raw'),
        ],
    TCP: [
        ('src', 'H', 0),
        ('dst', 'H', 2),
        ('seqnum', 'I', 4),
        ('acknum', 'I', 8),
        ('dataoffs', 'B', 12),
        ('flags', 'B', 13),
        ('winsize', 'H', 14),
        ('checksum', 'H', 16),
        ('urgent', 'H', 18),
        ('hsize', lambda x:x.dataoffs >> 2, 'self'),
        ('_str', '%(src)s->%(dst)s', 'raw'),
        ],
    UDP: [
        ('src', 'H', 0),
        ('dst', 'H', 2),
        ('size', 'H', 4),
        ('checksum', 'H', 6),
        ('hsize', 8, 'raw'),
        ('_str', '%(src)s->%(dst)s', 'raw'),
        ],
    }



if __name__ == '__main__':
    import math
    import sys
    import time
    for i, p in enumerate(PCap(sys.argv[1]).packets):
        print(f'#{p.num} @ {time.strftime("%Y.%m.%d,%H:%M:%S", time.gmtime(p.ts))}.{p.ts_us // 1000:03d}:', p)



"""
iogp.gui.qt: GUI wrappers and helper functions for PySide6 (Qt5).

Author: Vlad Topan (vtopan/gmail)
"""
import fnmatch
from logging import error, warn, info
import re

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import (SIGNAL, QByteArray, QCoreApplication, QDate, QDateTime,  # noqa: F401
        QEvent, QSortFilterProxyModel, Qt, QThread, Signal, Slot, QRect, QItemSelection,
        QItemSelectionModel)
from PySide6.QtGui import (QBrush, QClipboard, QColor, QCursor, QDesktopServices,  # noqa: F401
        QFont, QFontMetrics, QIcon, QIntValidator, QKeySequence, QPalette, QPixmap, QShortcut,
        QStandardItem, QStandardItemModel, QTextCursor, QTextOption, QAction, QTextCharFormat)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QCheckBox, QComboBox,  # noqa: F401
        QCompleter, QDateEdit, QDateTimeEdit, QFileDialog, QFrame, QGridLayout, QHBoxLayout,
        QHeaderView, QLabel, QLayout, QLineEdit, QListWidget, QMainWindow, QMessageBox,
        QPlainTextEdit, QPushButton, QScrollBar, QStatusBar, QStyleFactory, QTableView,
        QTableWidget, QTableWidgetItem, QTabWidget, QTextBrowser, QTextEdit, QTreeWidget,
        QTreeWidgetItem, QBoxLayout, QVBoxLayout, QWidget, QMenu, QMenuBar)

from ..data import AttrDict



class WidgetManager:
    """
    Widget manager for PySide6/Qt5.
    """

    def __init__(self, parent, widgets=None, auto_save_fields=None):
        self.widgets = None
        self.parent = parent
        self.auto_save_fields = auto_save_fields
        if widgets:
            self.load_widgets(widgets)

    def load_widgets(self, widgets):
        """
        Converts recursive structures of lists, dicts and GUI widgets to a proper PySide6 layout.

        A list is converted to a QVBoxLayout, a dict to a QHBoxLayout, everything else is presumed
        to be a widget and simply added to the current layout. The keys in each dict are mapped to
        the corresponding widgets in the output `.widgets` dict.

        Known keywords:
        - '!tab' converts a dict to a QTabWidget
        - '!vertical' converts a dict to a QVBoxLayout instead of a QHBoxLayout
        - '!horizontal' converts a list to a QHBoxLayout instead of a QVBoxLayout
        - '!border' creates a border (QFrame) around the layout (the value can be CSS styling)
        - '!stretch' sets stretching rules (e.g. '!stretch': (0, 2) sets the stretch property of
            widget 0 to 2; can be a list of tuples to set the stretch for multiple widgets

        :return: The top-level layout.
        """
        self.layout = None
        self.widgets = {}
        self.tabs = []
        if isinstance(widgets, str):
            # todo parse as JSON
            raise NotImplementedError(':(')
        if isinstance(widgets, (dict, list)):
            self.layout = self.create_layouts(widgets)
        else:
            raise ValueError(
                    f'Cannot process {widgets} of type {type(widgets)}, should be dict or JSON str!'
            )

    def create_layouts(self, widgets, name=None):
        """
        Parse the nested widget structure and generate the layout tree.
        """

        def _handle_directive(name, value=None):
            nonlocal layout
            if name.startswith('!stretch'):
                if not isinstance(value, list):
                    value = [value]
                for e in value:
                    layout.setStretch(*e)
            elif name == '!border':
                panel.setFrameStyle(QFrame.StyledPanel)
                panel.setStyleSheet('QFrame { margin: 0px; border: 1px solid #888; padding: 1px; }')
            elif name.startswith('!max_'):
                what = name.split('_')[1].title()
                if what not in ('Width', 'Height', 'Size'):
                    raise ValueError(f'Unknown directive {name}!')
                getattr(panel, 'setMaximum' + what)(value)
            elif name in ('!horizontal', '!vertical', '!tab', '!panel'):
                pass
            else:
                raise ValueError(f'Unknown directive {name}!')

        if not isinstance(widgets, (list, dict)):
            raise ValueError('The toplevel node must be a list or a dict!')
        if '!' in widgets:
            # a single '!' key can hold a sequence of directives
            widgets.update({e: ... for e in widgets.pop('!')})
        panel = None
        if '!panel' in widgets or '!max_width' in widgets or '!border' in widgets:
            panel = QFrame()
            panel.setStyleSheet('QFrame { margin: 0px; border: 0px; padding: 0px; }')
        layout = QVBoxLayout() if ((isinstance(widgets, list) and '!horizontal' not in widgets)
                or '!vertical' in widgets) else QHBoxLayout()
        tab = None
        if '!tab' in widgets:
            tab = QTabWidget()
            self.tabs.append(tab)
            self.widgets[name] = tab
            layout.addWidget(tab)
        if tab and not isinstance(widgets, dict):
            raise ValueError('Tab widgets must be dicts!')
        wsettings = {}
        for widget in widgets:
            if widget == '!tab':
                continue
            w_name = None
            add_layout_fun = layout.addLayout
            if isinstance(widgets, dict):
                w_name, widget = widget, widgets[widget]
            # handle directives
            if isinstance(w_name, str) and w_name[0] == '!':
                _handle_directive(w_name, widget)
                continue
            elif isinstance(widget, str):
                _handle_directive(widget)
                continue
            # add this widget(s) or layout
            if tab:
                tab_widget = QWidget()
                tab.addTab(tab_widget, w_name)
                add_layout_fun = tab_widget.setLayout
            if isinstance(widget, (list, dict)):
                widget = self.create_layouts(widget, name=w_name)
                add_layout_fun(widget)
            elif isinstance(widget, QLayout):
                add_layout_fun(widget)
            elif hasattr(widget, 'parent_layout'):   # VLabelEdit
                add_layout_fun(widget.parent_layout)
            else:
                if tab:
                    tab_layout = QVBoxLayout()
                    tab_widget.setLayout(tab_layout)
                    tab_layout.addWidget(widget)
                else:
                    layout.addWidget(widget)
                for k, v in wsettings.items():
                    getattr(widget, k)(v)
            prev = self.widgets.get(w_name, None)
            if (not prev) or isinstance(prev, (QHBoxLayout, QVBoxLayout, QGridLayout)):
                self.widgets[w_name] = widget
                widget.setObjectName(w_name)
        if panel:
            panel.setLayout(layout)
            layout = QHBoxLayout()
            layout.addWidget(panel)
        return layout

    def connect_events(self):
        """
        Connect self.parent.<btn_widget_name>_clicked handlers to .clicked events and others.
        """
        for k, v in self.widgets.items():
            for cls, signal, suffix, warn_if_missing in [
                    (QPushButton, 'clicked', 'clicked', True),
                    (QLineEdit, 'returnPressed', 'pressed_enter', False),
                    (QComboBox, 'currentTextChanged', 'changed', False),
                    (QCheckBox, 'stateChanged', 'changed', False),
                    ]:
                if isinstance(v, cls):
                    name = k + '_' + suffix
                    method = getattr(self.parent, name, None)
                    if method:
                        getattr(self[k], signal).connect(method)
                    elif warn_if_missing:
                        warn(f'Missing .{name}()!')

    def save_to_cfg(self, cfg, fields=None, keys=('history', 'field_values')):
        """
        Auto-save GUI field values to the given subkey of cfg.
        """
        fields = fields or self.auto_save_fields or []
        for k in keys:
            if k not in cfg:
                cfg[k] = AttrDict()
            cfg = cfg[k]
        for field in fields:
            lst = fnmatch.filter(self.wmap, field) if '*' in field else [field]
            for k in lst:
                widget = self[k]
                for (wtype, meth) in [
                        (QCheckBox, 'isChecked'),
                        (QComboBox, 'currentText'),
                        (QListWidget, lambda x: [x.item(i).text() for i in range(x.count())]),
                        (QTextBrowser, 'toHtml'),
                        ]:
                    if isinstance(widget, wtype):
                        val = getattr(widget, meth)() if isinstance(meth, str) else meth(widget)
                        break
                else:
                    val = widget.text()
                cfg[k] = val

    def load_from_cfg(self, cfg, keys=('history', 'field_values')):
        """
        Load widget contents from the given subkeys of cfg.
        """
        for k in keys:
            if k not in cfg:
                cfg[k] = AttrDict()
            cfg = cfg[k]
        for k, v in list(cfg.items()):
            if v:
                if not hasattr(self, k):
                    del cfg[k]
                    continue
                widget = self[k]
                for (wtype, meth) in [
                        (QCheckBox, 'setChecked'),
                        (QComboBox, 'setCurrentText'),
                        (QListWidget, 'addItems'),
                        (QTextBrowser, 'setHtml'),
                        ]:
                    if isinstance(widget, wtype):
                        getattr(widget, meth)(v)
                        break
                else:
                    widget.setText(v)

    def __getattr__(self, name):
        if name in self.widgets:
            return self.widgets[name]
        super().__getattribute__(name)

    def __contains__(self, name):
        return name in self.widgets

    def __getitem__(self, name):
        return self.widgets[name]



def apply_style(style, cfg, main_window, wm=None, additional_colors=None, custom_css=None):
    """
    Apply a style from cfg.path('stylesheets')
    """
    if wm is None:
        wm = main_window.wm
    cfg.style.colors = colors = AttrDict(Background='#F8F8F0', Text='#111111', Grid='#BBBBBB',
            Line='#111111', link='initial')
    if style == 'Default':
        css = ''
    else:
        try:
            filename = f'{cfg.path("stylesheets")}/{style}.style'
            info(f'Loading style from {filename}...')
            for line in open(filename):
                line = line.strip()
                if not line or line[0] == '#':
                    continue
                k, v = re.search(r'(^[^:]+):\s+(.+)$', line).groups()
                colors[k] = v
        except Exception as e:
            return error(f'Failed loading style from {filename}: <{e.__class__.__name__}> {e}')
        css_tree = {
            'QCheckBox::indicator': ['  border: 1px solid;'],
            'QLineEdit': ['  border: 1px solid;', '  padding: 0px;'],
            'QTabBar::tab': ['  border-width: 1px;', '  border-style: solid;',
                    '  padding: 2px 5px 2px 5px;', '  border-radius: 3px;', '  margin: 2px;'],
            'QToolTip': ['  border: 0;'],   # without this, the background color is not applied...
            'QTableWidget': ['  outline: 0;'],
            'QTableWidget::item': ['  border: 1px solid;'],
            'QTableWidget::selected': ['  border: 1px solid;'],
            # 'QPushButton': ['  border-radius: 3px;']
            }
        # todo: better styling, get non-color options from .style - ref https://doc.qt.io/qt-6/stylesheet-examples.html
        for field, maps_to in {
                'background': {
                    'background-color': ['*', 'QTabBar::tab', 'QCheckBox::indicator:unchecked',
                            'QHeaderView::section',     # for table (QTableWidget) headings
                        ],
                    'color': ['QCheckBox::indicator:checked'],
                    },
                'editable': {
                    'background-color': ['QLineEdit', 'QListWidget', 'QComboBox'],
                    },
                'text': {
                    'color': ['*'],
                    },
                'selection': {
                    'background-color': ['QTabBar::tab:selected', 'QListView::item:selected',
                            'QTableWidget::item:selected'
                        ],
                    },
                'lines': {
                    'border-color': ['*', 'QCheckBox::indicator', 'QLineEdit', 'QListWidget',
                            'QTabBar::tab', 'QTabWidget', 'QCheckBox'
                        ],
                    },
                'hover': {
                    'background-color': [
                            'QToolTip',
                            'QCheckBox::indicator:indeterminate:hover',
                            'QCheckBox::indicator:pressed',
                            'QCheckBox::indicator:checked:hover',
                            'QCheckBox::indicator:unchecked:hover',
                            'QCheckBox::indicator:indeterminate:hover',
                            'QPushButton:hover', 'QLineEdit:hover', 'QTabBar:hover', 'QListView::item:hover',
                            'QTableWidget::item:hover',
                        ],
                    },
                'buttons': {
                    'background-color': ['QPushButton', 'QCheckBox:indicator'],
                    },
                'checkbox_image': {
                    'image': ['QCheckBox::indicator:checked'],
                    }
                }.items():
            if field not in colors:
                continue
            for name, cats in maps_to.items():
                for cat in cats:
                    if cat not in css_tree:
                        css_tree[cat] = []
                    css_tree[cat].append(f'  {name}: {colors[field]};')
        css = []
        for cat, values in sorted(css_tree.items()):
            css += [f'{cat} {{'] + sorted(values) + ['}\n']
        css = '\n'.join(css).replace('$img', cfg.path('images'))
    if custom_css:
        css += '\n\n' + custom_css
    open(f'{cfg.path("stylesheets")}/current.css', 'w').write(css)
    main_window.setStyleSheet(css)


# monkey-patch sensible methods for layouts
def add_to_layout(self, *things, **kwargs):
    for thing in things:
        if isinstance(thing, QLayout):
            self.addLayout(thing)
        else:
            self.addWidget(thing)
    for name, thing in kwargs.items():
        if isinstance(thing, QLayout):
            self.addLayout(thing)
        else:
            self.addWidget(thing)
        setattr(self, name, thing)


QHBoxLayout.add = add_to_layout
QVBoxLayout.add = add_to_layout


def menu_from_dict(win, menu, mdict):
    """
    Creates a menu based on a dict template.

    :param win: Window (Qt object with the `menu_...()` action handlers).
    """
    if not hasattr(win, 'menus'):
        win.menus = {}
    for k, v in mdict.items():
        if k.startswith('-'):
            menu.addSeparator()
        else:
            if type(v) is dict:
                submenu = QMenu(k)
                win.menus[k] = submenu
                menu.addMenu(submenu)
                menu_from_dict(win, submenu, v)
            else:
                handler = getattr(win, 'menu_' + re.sub(r'\W+', '', k), None)
                if handler:
                    action = menu.addAction(QAction(k, win, triggered=handler, shortcut=v))
                    win.menus[k] = action



"""
iogp.data: Various data processing tools.

Author: Vlad Topan (vtopan/gmail)
"""
from .fileid import get_file_type, get_ft_category, get_ft_ext, FT_UNKNOWN, FT_EMPTY, FT_CATEGORY
from .carve import extract_embedded_files
from .ds import AttrDict, Config, dict_to_AttrDict
from .archive import Archive
from .str import (ellipsis, parse_num, parse_percent, format_date, format_datetime, format_num,
        format_percent, format_size, int_to_bytes, bytes_to_int, decode_cp437, clean_filename,
        UNICODE_TO_CP437, CP437_TO_UNICODE)
from .dataview import DataView



"""
iogp.db.sqlite: SQLITE wrapper implementing the GenDBObj interface in `iogp.dbobj`.

Author: Vlad Topan (vtopan/gmail)
"""

from .dbobj import GenDBObj

import re
import ctypes
import time
import zlib
import io
import sqlite3


SQLITE_OK = 0
SQLITE_ERROR = 1
SQLITE_BUSY = 5
SQLITE_LOCKED = 6

SQLITE_OPEN_READONLY = 1
SQLITE_OPEN_READWRITE = 2
SQLITE_OPEN_CREATE = 4


DEF_ID_KEY = 'id'
DEF_INT_KEY = 'rowid'
ISYM = '?'
DEF_PK_TYPE = 'INT PRIMARY KEY ASC'



class DB(GenDBObj):
    """
    SQLITE DB object.
    """


    def __init__(self, path=None, serialized=False, **kwargs):
        super().__init__(serialized=serialized, **kwargs)
        self.DEF_ID_KEY = DEF_ID_KEY
        self.DEF_INT_KEY = DEF_INT_KEY
        self.ISYM = ISYM
        self.DEF_PK_TYPE = DEF_PK_TYPE
        self._con_args = kwargs
        if path:
            self.connect(path)


    def connect(self, path, fast=0, in_memory=0, isolation=None, journal=None, synchronous=None):
        """
        Connect to a DB.

        in_memory=1 causes the DB to be kept in memory; run .save_to_disk() periodically.
        isolation=None means autocommit. Set to '' for default behavior (other values: DEFERRED,
        IMMEDIATE or EXCLUSIVE).
        """
        t0 = time.time()
        self.in_memory = in_memory
        self.fast = fast
        self.journal = journal
        self.synchronous = synchronous
        self.isolation = isolation
        if self.in_memory:
            self.fast = 1
            self.path = ':memory:'
            self.realpath = path
        else:
            self.realpath = self.path = path
        if self.fast:
            if not self.journal:
                self.journal = 'OFF'
            if not self.synchronous:
                self.synchronous = 'OFF'
            if self.isolation is None:
                self.isolation = 'EXCLUSIVE'
        self.conn = sqlite3.connect(self.path, isolation_level=isolation, **self._con_args)
        # if self.in_memory:
        #    data = zlib.decompress(open(self.realpath), 'rb').read()
        self.conn.row_factory = sqlite3.Row
        self.conn.text_factory = str
        if journal:
            self.conn.execute("PRAGMA journal_mode = %s" % journal)
        if synchronous:
            self.conn.execute("PRAGMA synchronous = %s" % synchronous)
        self.cur = self.conn.cursor()
        self.connect_time = time.time() - t0
        self.dbg('DB loaded from %s in %.2fs.' % (self.realpath, self.connect_time))


    def close(self):
        if self.in_memory:
            self.save_to_disk()
        self.conn.close()


    def save_to_disk(self):
        """
        This is only needed if 'fast=1' was passed to commit (and the DB is actually kept in
        memory.
        """
        if self.in_memory:
            self.save_to_file()
        else:
            raise ValueError('save_to_disk() called for disk database!')


    def save_to_file(self, sql_filename=None, db_filename=None):
        """ Save a database to a file as SQL. Works on the current DB if none given. """
        if db_filename:
            conn = sqlite3.connect(db_filename, self._con_args)
        else:
            conn = self.conn
        sio = io.StringIO()
        for line in conn.iterdump():
            sio.write(line)
        conn.close()
        sio.seek(0)
        data = zlib.compress(sio.read())
        open(sql_filename or self.realpath, 'wb').write(data)


    def add_explicit_idpk(self, table):
        sql = self.get_table_sql(table)
        if 'primary key' in sql.lower():
            self.log('Table [%s] already has an explicit primary key: %s!' % (table, sql), 'err')
            return
        columns = ', '.join(self.get_table_columns(table))
        for q in [
                'DROP TABLE IF EXISTS $t_new_',
                'CREATE TABLE $t_new_ (id INTEGER PRIMARY KEY ASC, ' + sql.split('(', 1)[1],
                'INSERT INTO $t_new_(id, %s) SELECT rowid, %s FROM $t' % (columns, columns),
                'ALTER TABLE $t RENAME TO $t_old_',
                'ALTER TABLE $t_new_ RENAME TO $t',
                'DROP TABLE $t_old_',
                ]:
            self.execute(q.replace('$t', table))


    def vacuum(self):
        self.execute('VACUUM;')


    def table_exists(self, table):
        return bool(self.fetchone("SELECT name FROM sqlite_master WHERE type = 'table' AND name = ?;", (table,)))


    def get_table_columns(self, table):
        ### UNSAFE INTERPOLATION!
        return [e['name'] for e in self.fetchall("PRAGMA table_info(%s)" % table)]


    def get_table_sql(self, table):
        return self.fetchone("SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = 'table'", (table,))[0]


    def dump_table_heads(self, tables, entries=20):
        for t in sorted(tables):
            cnt = self.fetchone("SELECT count(1) FROM %s" % t)[0]
            self.log((' %s (%s) ' % (t, cnt)).center(100, '*') + '\n')
            lst = self.fetchall("SELECT rowid, * FROM %s LIMIT %s" % (t, entries))
            if not lst:
                self.log('No entries!\n')
                continue
            keys = lst[0].keys()
            ms = dict(zip(keys, [(max(len(k), max(len(str(lst[i][k]).strip()) for i in range(len(lst)))) + 2)
                    for k in keys]))
            hdr = '| %s |' % (' | '.join('%-*s' % (ms[ee], ee.strip()) for ee in keys))
            sep_line = '*%s*' % ('-' * (len(hdr) - 2))
            self.log(sep_line)
            self.log(hdr)
            self.log(sep_line)
            for e in lst:
                self.log('| %s |' % (' | '.join('%-*s' % (ms[ee], str(e[ee])[:50].strip()) for ee in keys)))
            self.log(sep_line)
            self.log('\n')


    def create_tables(self, tables):
        """
        Creates tables (if they don't already exist). 'tables' is a dict mapping table names to a
        valid SQL table description. E.g:

        db.create_tables({'t1':'id INTEGER PRIMARY KEY ASC, name TEXT NOT NULL',})
        """
        for table in tables:
            if not self.table_exists(table):
                self.execute("CREATE TABLE %s(%s)" % (table, tables[table].replace('${PK}', DEF_PK_TYPE)))
                self.conn.commit()


    def create_indexes(self, tables):
        """
        Ensures indexes exist on the given columns for the given tables. 'tables' is a dict mapping
        table names to a list of column names (strings).
        """
        for table in tables:
            for column in tables[table]:
                self.execute("CREATE INDEX IF NOT EXISTS index_%s ON %s (%s)" % (re.sub('[, ]+', '_',
                        column), table, column))



def backup(src_path, dest_path):
    """
    Create a backup of an sqlite database (ctypes based).
    """
    p_src_db = ctypes.c_void_p(None)
    p_dst_db = ctypes.c_void_p(None)
    null_ptr = ctypes.c_void_p(None)

    if not sqlite_dll:
        init()

    ret = sqlite_dll.sqlite3_open_v2(src_path, ctypes.byref(p_src_db), SQLITE_OPEN_READONLY, null_ptr)
    assert ret == SQLITE_OK
    assert p_src_db.value is not None
    ret = sqlite_dll.sqlite3_open_v2(dest_path, ctypes.byref(p_dst_db), SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,
            null_ptr)
    assert ret == SQLITE_OK
    assert p_dst_db.value is not None

    p_backup = sqlite_dll.sqlite3_backup_init(p_dst_db, 'main', p_src_db, 'main')
    assert p_backup is not None

    while True:
        ret = sqlite_dll.sqlite3_backup_step(p_backup, 20)
        remaining = sqlite_dll.sqlite3_backup_remaining(p_backup)
        # pagecount = sqlite_dll.sqlite3_backup_pagecount(p_backup)
        if remaining == 0:
            break
        if ret in (SQLITE_OK, SQLITE_BUSY, SQLITE_LOCKED):
            sqlite_dll.sqlite3_sleep(100)

    sqlite_dll.sqlite3_backup_finish(p_backup)

    sqlite_dll.sqlite3_close(p_dst_db)
    sqlite_dll.sqlite3_close(p_src_db)



def init():
    """
    Initialize the ctypes wrapper over sqlite3.dll for the `backup()` function.
    """
    global sqlite_dll
    if not sqlite_dll:
        sqlite_dll = ctypes.CDLL('sqlite3.dll')
        sqlite_dll.sqlite3_backup_init.restype = ctypes.c_void_p


sqlite_dll = None



"""
iogp.gui.qthexed: Pure-Python PySide 6 / Qt5 hex viewer / editor.

Minimal example:

    # import random
    import sys
    # data = bytes(random.randint(0, 255) for i in range(1024))
    fn = sys.argv[-1]   # filename given as parameter or the script itself
    data = open(fn, 'rb').read()
    app = QtWidgets.QApplication()
    he = VHexEditor(data=data)
    he.show()
    sys.exit(app.exec_())

Author: Vlad Topan (vtopan/gmail)
"""
from .qt import (QWidget, Qt, SIGNAL, QScrollBar, QHBoxLayout, QVBoxLayout, QLabel,
        QShortcut, QStatusBar, QTableView, QHeaderView, QPlainTextEdit, QFont, QFontMetrics,
        QStandardItemModel, QColor, QBrush, QStandardItem, QTextOption, QTextCursor,
        QTextCharFormat, QRect, QItemSelection, QItemSelectionModel)
from ..data import DataView, CP437_TO_UNICODE, decode_cp437, AttrDict


NONPRINTABLE = {k: CP437_TO_UNICODE[k] for k in (0, 7, 8, 9, 10, 13, 0x7F)}

STYLE = AttrDict({
    'even_text': '#800',
    'odd_text': '#008',
    'selection': '#66F',
    'background': '#FFF',
    'highlight': '#DD6',
    'font': 'Terminal',
    'font_size': 8,
})

# todo: more flexible styling
HEADER_CSS = '''
::section {
    border: 0px;
    padding: 1px;
    text-align: center;
    font: monospace;
    }
::section:checked {
    color: white;
    font-weight: normal;
    background-color: blue;
    }
::section:hover {
    color: blue;
    }
'''

TABLE_CSS = '''
::item {
    gridline-color: white;
    }
::item:focus {
    border: 0px;
    color: red;
    gridline-color: white;
    }
::item:hover {
    color: blue;
    }
'''

BINVIEW_CSS = '''
'''



class VBaseEditor(QWidget):
    """
    Base class for VHexEditor and VBinViewer.
    """
    def __init__(self, viewer, data=None, filename=None, readonly=True, statusbar=1, columns=16,
                min_width=800, style=None):
        super().__init__()
        self.viewer = viewer
        self.set_style(style)
        self._font = QFont(self.style.font, self.style.font_size)
        self.metrics = QFontMetrics(self._font)
        self.row_height = self.metrics.height() + 4
        if columns is None:
            columns = 80
        self.columns = columns
        self.rows = 1
        self.perpage = self.columns * self.rows
        self.scroll = QScrollBar(Qt.Vertical)
        self.scroll.setPageStep(1)
        self.scroll.valueChanged[int].connect(self.jump_to_row)
        self.l1 = QVBoxLayout()
        self.l2 = QHBoxLayout()
        self.l2.addWidget(self.viewer)
        self.l2.addWidget(self.scroll)
        self.l1.addLayout(self.l2)
        self.statusbar = statusbar and QStatusBar()
        if self.statusbar:
            self.l1.addWidget(self.statusbar)
            self.status = {}
            self.status['filename'] = QLabel('-')
            self.lbl_offset = QLabel('Offset:')
            self.status['offset'] = QLabel('-')
            self.statusbar.addWidget(self.status['filename'])
            self.statusbar.addWidget(self.lbl_offset)
            self.statusbar.addWidget(self.status['offset'])
        self.setLayout(self.l1)
        self.setMinimumWidth(min_width)
        self.dv = None
        self.own_dv = False
        self.open(data, filename, readonly)

    def set_style(self, style):
        """
        Set styling dict (see `STYLE` for keys).
        """
        self.style = AttrDict(STYLE)
        if style:
            # todo: use .style in currently-hardcoded locations
            self.style.update(style)

    def wheelEvent(self, evt):
        """
        Hooked to implement scrolling by mouse wheel.
        """
        self.jump_rows(1 if evt.angleDelta().y() > 0 else -1)
        evt.accept()

    def keyPressEvent(self, evt):
        """
        Hooked to implement scrolling by up/down/pgup/pgdn.

        Note: only events which aren't caught by self.viewer end up here.
        """
        key = evt.key()
        if key in (Qt.Key_PageUp, Qt.Key_PageDown):
            self.jump_pages(1 if key == Qt.Key_PageUp else -1)
            evt.accept()
        elif key in (Qt.Key_Up, Qt.Key_Down):
            self.jump_rows(1 if key == Qt.Key_Up else -1)
            evt.accept()

    def open(self, data=None, filename=None, readonly=True):
        """
        Open a file (or a data buffer).
        """
        self.offs = 0
        if self.dv is not None and self.own_dv:
            self.dv.close()
        if isinstance(data, DataView):
            self.dv = data
            self.own_dv = False
        else:
            self.dv = DataView(data=data, filename=filename, readonly=readonly)
            self.own_dv = True
        self.size = len(self.dv)
        self.scroll.setRange(0, self.size // self.columns)
        if self.statusbar:
            self.status['filename'].setText((f'File: <b>{filename}</b>')
                    if filename else (f'Raw data ({self.size} bytes)'))
        self.update()

    def jump_pages(self, pages=1):
        """
        Jump a number of pages up / down.

        :param pages: Number of pages to jump; negative values jump down.
        """
        self.jump(self.offs - pages * self.perpage)

    def jump_rows(self, rows=1):
        """
        Jump a number of rows up / down.

        :param rows: Number of rows to jump; negative values jump down.
        """
        self.jump(self.offs - rows * self.columns)

    def jump_to_row(self, row):
        """
        Jump to the given row id.
        """
        self.jump(row * self.columns)

    def update(self):
        """
        Update contents.
        """
        self.jump(force=True)



class VBinViewer(VBaseEditor):
    """
    Binary viewer widget.

    :param data: The raw data (bytes or bytearray).
    :param filename: The file name containing the data.
    :param columns: Number of columns (default: fit window).
    """

    def __init__(self, data=None, filename=None, statusbar=1, columns=None, min_width=800,
                style=None):
        super().__init__(QPlainTextEdit(),
                data=data, filename=filename, readonly=True, statusbar=statusbar,
                columns=columns, min_width=min_width, style=style)
        self.viewer.setWordWrapMode(QTextOption.WrapAnywhere)
        self.viewer.setFont(self._font)
        self.update()

    def sel_changed(self, new, old):
        """
        Selection changing is monitored to allow highlighting matching selected bytes between the
        hex dump and the text.
        """

    def jump(self, offs=0, force=False):
        """
        Jump to an offset.
        """
        new_offs = min(max(0, offs), max(0, len(self.dv) - self.perpage))
        new_offs = new_offs // self.columns * self.columns
        # print('JUMP:', new_offs, self.offs, force)
        if (new_offs != self.offs) or force:
            self.offs = new_offs
            decoded = decode_cp437(self.current_data)
            decoded = '\n'.join(decoded[i:i + self.columns] for i in range(0, len(decoded),
                    self.columns))
            self.viewer.setPlainText(decoded)

    def resizeEvent(self, evt):
        """
        This keeps self.rows/columns/perpage updated and regenerates the self.viewer content.
        """
        rect = self.viewer.fontMetrics().boundingRect('W')
        self.rows = (self.viewer.height() - 10) // rect.height()
        self.columns = (self.viewer.width() - 30) // rect.width()
        self.perpage = self.columns * self.rows
        self.scroll.setRange(0, self.size // self.columns)
        self.update()
        super().resizeEvent(evt)

    def adjust_pos_from_gui(self, pos):
        """
        Compute file offset from GUI / selection offset (subtract newlines, add offs).
        """
        res = pos - pos // self.columns + self.offs
        return res

    def adjust_pos_to_gui(self, pos):
        """
        Compute GUI offset from file offset (add newlines, subtract offs).
        """
        res = pos - self.offs
        res += res // self.columns
        return res

    @property
    def current_data(self):
        """
        The currently-displayed data.
        """
        return self.dv[self.offs:self.offs + self.perpage]

    @property
    def sel_start(self):
        """
        Selection start.
        """
        pos = self.viewer.textCursor().selectionStart()
        print('SELSTART:', pos)
        if pos is not None:
            pos = self.adjust_pos_from_gui(pos)
        return pos

    @property
    def sel_end(self):
        """
        Selection end.
        """
        pos = self.viewer.textCursor().selectionEnd()
        if pos is not None:
            pos = self.adjust_pos_from_gui(pos)
        return pos

    @property
    def sel_length(self):
        """
        Selection length.
        """
        tc = self.viewer.textCursor()
        return tc.selectionEnd() - tc.selectionStart()

    def select(self, start, end, jump=True):
        """
        Select text by offset.
        """
        if jump:
            self.jump(max(0, start - 2 * self.columns))
        start = self.adjust_pos_to_gui(start)
        end = self.adjust_pos_to_gui(end)
        cur = self.viewer.textCursor()
        # print('moving selection: setpos:', start)
        cur.setPosition(start)
        cur.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, n=end - start)
        self.viewer.setTextCursor(cur)

    def highlight(self, ranges):
        """
        Highlight a list of (start, end) ranges.
        """
        cursor = QTextCursor(self.document())
        for start, end in ranges:
            # keep track of newlines
            cursor.setPosition(start + start // self.columns)
            cursor.setPosition(end + end // self.columns, QTextCursor.KeepAnchor)
            format = QTextCharFormat()
            format.setBackground(QColor(self.style.highlight))
            cursor.mergeCharFormat(format)

    def __getattr__(self, attr):
        """
        Forward some methods.
        """
        if attr in ('document',):
            return getattr(self.viewer, attr)
        raise AttributeError(f'Invalid attribute: {attr}!')



class VHexEditor(VBaseEditor):
    """
    Hex editor widget.

    :param data: The raw data (bytes or bytearray).
    :param filename: The file name containing the data.
    """

    def __init__(self, data=None, filename=None, readonly=True, statusbar=1, columns=16,
                min_width=800, style=None):
        super().__init__(QTableView(showGrid=0, styleSheet=TABLE_CSS),
                data=data, filename=filename, readonly=readonly, statusbar=statusbar,
                columns=16, min_width=min_width, style=style)
        self.viewer.setFont(self._font)
        self.dm = QStandardItemModel(self.rows, self.columns)
        self.viewer.setModel(self.dm)
        self.viewer.horizontalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.viewer.horizontalHeader().setHighlightSections(1)
        self.viewer.verticalHeader().setHighlightSections(1)
        self.viewer.horizontalHeader().setStretchLastSection(0)
        self.viewer.selectionModel().selectionChanged.connect(self.sel_changed)
        self.update()

    def sel_changed(self, new, old):
        """
        Selection changing is monitored to allow highlighting matching selected bytes between the
        hex dump and the text.
        """
        new, old = ([(e.row(),
                (e.column() + self.columns) % (2 * self.columns))
                for e in lst.indexes()] for lst in (new, old))
        for e in old:
            if e in new:
                continue
            if item := self.dm.item(*e):
                item.setBackground(QColor(self.style.background))
        for e in new:
            if e in old:
                continue
            if item := self.dm.item(*e):
                item.setBackground(QColor(self.style.selection))
        if self.statusbar:
            idx = self.viewer.selectionModel().currentIndex()
            offs = self.offs + idx.row() * self.columns + idx.column()
            self.status['offset'].setText('<b>0x%X</b> (%d)' % (offs, offs))

    def select(self, start, count, jump=True):
        """
        Select a range of bytes.
        """
        base_row = max(0, (start // self.columns) - 1)
        if jump:
            self.jump_to_row(base_row)
        base_row = self.offs // self.columns
        selmdl = self.viewer.selectionModel()
        mdl = self.viewer.model()
        selmdl.clearSelection()
        max_rows = self.rows
        # todo: use QTextCursor.NextCell?
        while count > 0 and max_rows >= 0:
            row = start // self.columns - base_row
            col1 = start % self.columns
            col2 = min(self.columns - 1, col1 + count - 1)
            selected = col2 - col1 + 1
            count -= selected
            start += selected
            # todo: use row selection when support is added
            #if selected == self.columns:
            #    selmdl.select(mdl.index(row, 0), QItemSelectionModel.Select | QItemSelectionModel.Rows)
            #else:
            selmdl.select(QItemSelection(mdl.index(row, col1), mdl.index(row, col2)), QItemSelectionModel.Select)
            max_rows -= 1

    def resizeEvent(self, evt):
        """
        This keeps self.rows/columns/perpage updated and regenerates the self.viewer content.
        """
        self.rows = (self.viewer.height()
                - self.viewer.verticalHeader().height() - 6) // self.row_height
        self.perpage = self.columns * self.rows
        self.dm.setRowCount(self.rows)
        self.scroll.setRange(0, self.size // self.columns)
        self.jump(self.offs)
        super().resizeEvent(evt)

    @property
    def sel_start(self):
        """
        Selection start.
        """
        idx = self.viewer.selectionModel().currentIndex()
        if col := self.column() >= self.columns + 1:
            col -= self.columns + 1
        else:
            col -= 1
        return idx.row() * self.columns + max(0, col)

    @property
    def sel_end(self):
        """
        Selection end.
        """
        idx = self.viewer.selectionModel().selectedIndexes()[-1]
        if col := self.column() >= self.columns + 1:
            col -= self.columns + 1
        else:
            col -= 1
        return idx.row() * self.columns + max(0, col)

    def jump(self, offs=0, force=False):
        """
        Jump to the given offset (this is the main workhorse which updates self.viewer contents).
        """
        # todo: only update on `force` or of offs != self.offs
        brushes = [QBrush(QColor(x)) for x in (self.style.odd_text, self.style.even_text)]
        hv = self.viewer
        if not hv.selectionModel():
            return
        crt = hv.selectionModel().currentIndex()
        vh = hv.verticalHeader()
        hh = hv.horizontalHeader()
        # ## get data window
        self.offs = max(0, min(offs, self.size - self.perpage))
        data = self.dv[self.offs:self.offs + self.perpage]
        # ## generate & set labels
        addrlabels = ['%08X' % i for i in range(self.offs, self.offs + self.perpage, self.columns)]
        self.dm.setVerticalHeaderLabels(addrlabels)
        self.dm.setHorizontalHeaderLabels([str(x) for x in range(self.columns)]
                + [''] * self.columns)
        # ## generate data for datamodel (actual viewer contents)
        idx = 0   # offset inside the data window (index in `data`)
        for row in range(self.rows):
            for col in range(self.columns):
                if idx < len(data):
                    c = data[idx]
                    hitem = QStandardItem('%02X' % c)
                    hitem.setTextAlignment(Qt.AlignCenter)
                    hitem.setForeground(brushes[col % 2])
                    titem = QStandardItem(bytes([
                            32 if c in NONPRINTABLE else c]).decode('cp437'))
                    titem.setTextAlignment(Qt.AlignCenter)
                    titem.setForeground(brushes[col % 2])
                else:
                    hitem = titem = None
                self.dm.setItem(row, col, hitem)
                self.dm.setItem(row, col + self.columns, titem)
                idx += 1
        # ## format view
        hh.setDefaultSectionSize(hv.fontMetrics().horizontalAdvance('A') + 2)
        hh.setMaximumSectionSize(hv.fontMetrics().horizontalAdvance('AAA'))
        for i in range(self.columns):
            hh.resizeSection(i, hv.fontMetrics().horizontalAdvance('A') + 2)
        vh.setDefaultSectionSize(self.row_height)
        vh.setDefaultAlignment(Qt.AlignCenter)
        vh.setStyleSheet(HEADER_CSS)
        hh.setStyleSheet(HEADER_CSS)
        # ## restore pre-scroll selection
        hv.setCurrentIndex(crt)
